# -----------------------------------
# Imports
# -----------------------------------

import numpy as np
from typing import List

from muphyn.packages.core.utils.log_manager import LogManager

# -----------------------------------
# Functions
# -----------------------------------


def _init_high_pass_filter(box, simulation_params) -> None:
    cut_pulstation = box['cut_frequency'] * 2 * np.pi

    box['num_vect'] = [1, 0]
    box['num_vect'].reverse()
    box['num_order'] = 1
    box['last_u_derivatives'] = [0 for _ in box['num_vect']]

    box['denom_vect'] = [1, cut_pulstation]
    box['denom_vect'].reverse()
    box['denom_order'] = 1
    box['last_y_derivatives'] = [0 for _ in box['denom_vect']]

    box['order'] = 1
    box['order_coeff'] = 1

    box['last_y'] = 0
    box['last_timing'] = -1


def _function_high_pass_filter(box, event_) -> List:
    
    v = 0
    current_u_derivatives = []
    current_y_derivatives = []
    input_value = box.get_input(0).value

    if box['last_timing'] == event_.timing :
        return []
    box['last_timing'] = event_.timing

    for j, coeff_u in enumerate(box['num_vect']) : 

        if j == 0 : 
            v += coeff_u * input_value
            current_u_derivatives.append(input_value)
        
        else :
            derivative = (current_u_derivatives[j - 1] - box['last_u_derivatives'][j - 1]) / event_.step_time
            v += coeff_u * derivative
            current_u_derivatives.append(derivative)

    if box['order'] > 0 :

        for j in range(box['order']) :

            coeff_y = box['denom_vect'][j]

            if j == 0 :
                v -= coeff_y * box['last_y']
                current_y_derivatives.append(box['last_y'])

            else :
                derivative = (current_y_derivatives[j - 1] - box['last_y_derivatives'][j - 1]) / event_.step_time
                v -= coeff_y * derivative
                current_y_derivatives.append(derivative)

        v /= box['order_coeff']

        for j in range(box['order']).__reversed__() :
            LogManager().info('time :', event_.timing, ' - order :', j, ' - derivative :', current_y_derivatives[j])
            v *= event_.step_time
            v += current_y_derivatives[j]

    else :
        v /= box['order_coeff']

    box['last_u_derivatives'] = current_u_derivatives
    box['last_y_derivatives'] = current_y_derivatives
    box['last_y'] = v

    """
    while (len(box['last_y_s']) > box['denom_order']) :
        del box['last_y_s'][0]
    """


    events: List = []

    for output in box.outputs :
        events.append(box.construct_signal_event(output, v))

    return events



#-----------------------------------
# Imports
#-----------------------------------

import time

import matplotlib.pylab as plt
from muphyn.packages.core.box_library.plci_core_boxes_libraries import BoxesLibraries
from muphyn.packages.core.plci_core_scheduler import Scheduler
from muphyn.packages.core.plci_core_scheduler_exception import SchedulerException
from muphyn.packages.core.scheduler_library.plci_core_schedulers_libraries import SchedulersLibraries

#-----------------------------------
# Methods
#-----------------------------------

def get_boxes_libraries () -> BoxesLibraries :
    print("Load boxes libraries")
    boxLibrary = BoxesLibraries()
    boxLibrary.add_library("D:/OneDrive - Ecole/OneDrive - Haute Ecole Louvain en Hainaut/2021-22 MaGe2/Stage/python-low-code-interface/box_library")
    boxLibrary.add_library("D:/OneDrive - Ecole/OneDrive - Haute Ecole Louvain en Hainaut/2021-22 MaGe2/Stage/python-low-code-interface/user_box")
    boxLibrary.load_libraries()
    return boxLibrary

def get_schedulers_library () -> SchedulersLibraries :
    print("Load schedulers libraries")
    schedulerLibrary = SchedulersLibraries()
    schedulerLibrary.add_library("D:/OneDrive - Ecole/OneDrive - Haute Ecole Louvain en Hainaut/2021-22 MaGe2/Stage/python-low-code-interface/scheduler_library")
    schedulerLibrary.load_libraries()
    return schedulerLibrary

def simulate (scheduler : Scheduler) -> bool :
        
    print("Start of the simulation #boxes : ", scheduler.diagram._boxes.__len__())
    start_time = time.time()
    schedulerException : SchedulerException = scheduler.schedule()
    stop_time = time.time()

    if schedulerException is None :
        print("The simulation is correctly finished and it took : ", stop_time - start_time, " s.")

        plt.show()

        return True

    else :
        print("The simulation encoutered an error after ", stop_time - start_time, " s of simulation.")
        schedulerException.print()
        return False



from muphyn.packages.core.plci_core_scheduler import Scheduler
from muphyn.packages.core.simulation_files.plci_core_simulations_importer import SimulationsImporter
from .unit_test import get_boxes_libraries, get_schedulers_library, simulate

# Get library
boxesLibraries = get_boxes_libraries()
schedulersLibraries = get_schedulers_library()
simulationImporter = SimulationsImporter()
simulation : Scheduler = simulationImporter.import_simulation(
                                'D:/OneDrive - Ecole/OneDrive - Haute Ecole Louvain en Hainaut/2021-22 MaGe2/Stage/python-low-code-interface/simulation/retroact.yaml',
                                schedulersLibraries, 
                                boxesLibraries)

if simulation is None :
    print('The simulation imported has not been loaded.')

else :
    simulate(simulation)



from muphyn.packages.interface.shapes.basic_shape.abstract_shape import AbstractShape
from muphyn.packages.interface.shapes.basic_shape.grouped_shapes import GroupedShapes
from muphyn.packages.interface.shapes.basic_shape.image_shapes import AbstractImage, Image, SVGImage, ImageShapeFactory
from muphyn.packages.interface.shapes.basic_shape.shapes import Circle, Line, Path, Polygon, Rectangle, Square, Text
from muphyn.packages.interface.shapes.basic_shape.layout_shapes import VerticalLayout


# General import
from typing import Optional, Type

# PyQt import
from PyQt5.QtCore import QPoint, QPointF
from PyQt5.QtWidgets import QWidget

# Project import
from muphyn.packages.interface.editors.multiphysics_model import *
from muphyn.packages.core.simulation.model_parser import OpenModelicaModelParser, OpenModelicaComponent, OpenModelicaConnection

class OpenModelicaMultiphysicsModel(AbstractMultiPhysicsModel):

    # Component look up table
    ComponentLUT: dict = {
        "Modelica.Electrical.Analog.Basic.Ground": GroundModel,
        "Modelica.Electrical.Analog.Basic.Resistor": ResistorModel,
        "Modelica.Electrical.Analog.Basic.VariableResistor": VariableResistorModel,
        "Modelica.Electrical.Analog.Sources.SignalCurrent": SignalCurrentSourceModel
    }

    def __init__(self, omParser: OpenModelicaModelParser, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)

        # Add components to view
        for component in omParser.components:
            self.addComponent(component)

        # Add connections to view
        for connection in omParser.connections:
            self.addConnection(connection)

    def addComponent(self, newOpenModelicaComponent: OpenModelicaComponent):

        # Build component model
        componentModel: AbstractComponentModel = OpenModelicaMultiphysicsModel.componentModelFactory(newOpenModelicaComponent)

        # Place item above
        componentModel.setZValue(1)

        # Add component to scene
        return super().addItem(componentModel)

    def addConnection(self, newOpenModelicaConnection: OpenModelicaConnection):
        # Convert open modelica connection points
        points = [QPointF(point.x(), -point.y()+20) * 4 for point in newOpenModelicaConnection.points()]

        # Build connection model
        connectionModel = ConnectionModel(points)

        # Place item under
        connectionModel.setZValue(0)

        # Add connection to scene
        return super().addItem(connectionModel)

    @staticmethod
    def componentModelFactory(openModelicaComponent: OpenModelicaComponent):
        newComponentModelType: Type[AbstractComponentModel] = OpenModelicaMultiphysicsModel.ComponentLUT[openModelicaComponent.library]

        # Get component data
        componentPosition: QPoint = openModelicaComponent.topLeft() * 4
        componentPosition = QPointF(componentPosition.x(), -componentPosition.y())
        componentRotation = -openModelicaComponent.rotation()

        newComponentModel = newComponentModelType(group_position=componentPosition)
        newComponentModel.setRotation(componentRotation)

        return newComponentModel


from typing import List
import numpy as np
import matplotlib.pyplot as plt



def derivate (n_minus_1 : float, n_0 : float, dT : float) -> float :
    return (n_0 - n_minus_1) / dT

def double_precision_derivate (n_minus_2 : float, n_minus_1 : float, n_0 : float, dT : float) -> float :
    ...

def derivate_second (n_minus_2 : float, n_minus_1 : float, n_0 : float, dT : float) -> float :
    return (n_0 - 2*n_minus_1 + n_minus_2) / (dT*dT)

def derivation_first (data : List[float], step : float, start_value : float = 0) -> List[float] :
    
    data_n_minus_1 = start_value
    vector = []

    for data_n_0 in data : 

        if len(vector) < 1 :
            vector.append(0)
            data_n_minus_1 = data_n_0
            continue
        
        y = derivate(data_n_minus_1, data_n_0, step)
        vector.append(y)

        data_n_minus_1 = data_n_0

    return vector

def derivation_second (data : List[float], step : float, start_values : List[float]) -> List[float] :
    
    data_n_minus_1 = start_values[0]
    data_n_minus_2 = start_values[1]

    vector = []

    for data_n_0 in data : 
    
        if len(vector) < 2 :
            vector.append(0)
            data_n_minus_2 = data_n_minus_1    
            data_n_minus_1 = data_n_0
            continue

        y = derivate_second(data_n_minus_2, data_n_minus_1, data_n_0, step)
        vector.append(y)

        data_n_minus_2 = data_n_minus_1    
        data_n_minus_1 = data_n_0

    return vector

order = 1
fs = 100.0       
cutoff = 30 

print('low pass filter')
print('order :', order)
print('fs :', fs)
print('cutoff :', cutoff)

# Creating the data for filteration
T = 1.0         # value taken in seconds
n = int(T * fs) # indicates total samples
t_a = np.linspace(0, T, n, endpoint=False)

data = []
for t in t_a :
    data.append(t * t * 5)

y_1 = derivation_first(data, 1/fs, (-1/fs)*(-1/fs)*5)
y_2 = derivation_second(data, 1/fs, [(-2/fs)*(-2/fs)*5, (-1/fs)*(-1/fs)*5])
y_3 = []
for i, y_1_ in enumerate(y_1) : 
    y_3.append(data[i] - y_1_ - y_2[i])

plt.figure(0)
plt.plot(t_a[2:], data[2:], 'r-', label='data')
plt.plot(t_a[2:], y_1[2:], 'g-', label='y1')
plt.plot(t_a[2:], y_2[2:], 'b-', label='y2')
plt.plot(t_a[2:], y_3[2:], 'c-', label='y3')
plt.xlabel('Time [sec]')
plt.grid()
plt.legend()

plt.show()


from os import system

import matplotlib.pyplot as plt

# =========== Simulation ===========

fs = 1000
dt = 1/fs
sim_time = 20
samples_number = int(fs * sim_time) + 1

_ = system('cls')
print('=============================================================')
print('||                                                         ||')
print('||                      Simulation                         ||')
print('||                                                         ||')
print('|| X = Y * 1 / (P² + P + 1)                                ||')
print('||                                                         ||')
print('|| - simulation time :', sim_time, '                                 ||')
print('|| - sample frequency :', fs, '                              ||')
print('|| - dt :', dt, '                                           ||')
print('|| - samples number :', samples_number, '                               ||')
print('||                                                         ||')
print('=============================================================')
print('')

# time
t = [n * dt for n in range(samples_number)]

# input
u = [1 for t_ in t]

# output
y = []

A = 2
B = 0
G = 0
K = 1

def test (u, a, b) :
    """
    # Sans coefficeitns (sous entendus, tous à 1)
    current_y = u
    current_y -= a
    current_y -= (a - b) / dt
    current_y *= dt
    current_y += (a - b) / dt
    current_y *= dt 
    current_y += a
    
    # Suite d'instructions.
    return current_y

    # Déjà distribué à la main.
    return ((u - a) * (dt ** 2)) - ((a - b) * dt) + 2 * a - b
    """
    # Avec coefficients.
    return ((K * (dt ** 2)) * u / A) - (G * (dt ** 2) * a / A) - ( B * (a - b) * dt / A) + (a - b) + a



for i, t_ in enumerate(t) :
#for i in range(10) :

    c_u = u[i - 1]
    c_a = 0
    c_b = 0

    if i >= 1 :
        c_a = y[i - 1]
    
    if i >= 2 :
        c_b = y[i - 2]

    temp_y = test(c_u, c_a, c_b)

    if (i == 0) :
        print(temp_y)

    y.append(temp_y)

print('at', (len(y) - 1) / fs, 's :', y[len(y) - 1])

# =========== Personnal notes ===========

print('')
print(' =====> Résultats <=====')
print('')
print('date : 18/02/2022')
print('')
print('Décomposition à la main pas encore complète.')
print('')
print('')
print('date : 21/02/2022')
print('')
print('Fin de la décomposition à la main.')
print('La courbe perçue est exactement la même que celle de xcos !!!!')
print('')
print('Ajout de coefficient devant les termes.')
print('Cela semble fonctionne de la même manière.')
print('')
print('Démarrage sur ces base d\'un nouveau test : EDT_SFT_0048_5')

# =========== Plot results ===========

t_renderable = t[0 : len(y)]
u_renderable = u[0 : len(y)]

plt.figure()
plt.plot(t_renderable, u_renderable, label="u(t)")
plt.plot(t_renderable, y, label="y(t)")
plt.legend()
plt.grid()
plt.xlabel('Time')
plt.show()


from PyQt5.QtWidgets import QDialog, QVBoxLayout

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar
from matplotlib.widgets import Cursor

from matplotlib.figure import Figure

class GraphWindow(QDialog):

    def __init__(self, parent=None , box = None):
        super().__init__(parent)

        self.figure = Figure()
        self.box = box

        # this is the Canvas Widget that displays the `figure`
        # it takes the `figure` instance as a parameter to __init__
        self.canvas = FigureCanvas(self.figure)

        # this is the Navigation widget
        # it takes the Canvas widget and a parent
        self.toolbar = NavigationToolbar(self.canvas, self)

        # set the layout
        layout = QVBoxLayout()
        self.setWindowTitle(str(self.box.index) + '-' +  str(self.box['title']))
        layout.addWidget(self.toolbar)
        layout.addWidget(self.canvas)
        layout.setStretch(1, 1)
        self.setLayout(layout)
        self.plot()

    def plot(self):
        data_y: dict = self.box['data_y']

        #fig = plt.figure(str(self.box.index) + '-' +  str(self.box['title'])) 
        self.figure.clear()
        
        
        ax = self.figure.add_subplot(111)
        ax.grid(True)
        
        ax.set_xlabel(self.box['t'])
        ax.set_ylabel(self.box['label_y'])

        for key, values in data_y.items() :
            ax.plot(self.box['data_x'], values, label=key)

        ax.legend(loc="upper right")

        cursor = Cursor(ax, horizOn=True, vertOn=True, useblit=True, color = 'r', linewidth = 1)
        self.annot = ax.annotate("", xy=(0,0), xytext=(-40,40),textcoords="offset points",
                    bbox=dict(boxstyle='round4', fc='linen',ec='k',lw=1),
                    arrowprops=dict(arrowstyle='-|>'))
        self.annot.set_visible(False)

        self.canvas.mpl_connect('button_press_event', self.onclick)
        self.canvas.draw()

    def onclick(self, event):
        i, x = _closest(event.xdata, self.box['data_x'])

        key = list(self.box['data_y'].keys())[0]
        y = self.box['data_y'][key][i]

        self.annot.xy = (x,y)
        text = "({:.2g}, {:.2g})".format(x,y)
        self.annot.set_text(text)
        self.annot.set_visible(True)
        self.canvas.draw()

def _closest(a:float, b:list):
    c = [abs(x-a) for x in b]
    i = c.index(min(c))
    return (i, b[i])


#-----------------------------------
# Imports
#-----------------------------------

from typing import List, Any

from PyQt5.QtCore import QCoreApplication, Qt
from PyQt5.QtWidgets import QListWidgetItem

from muphyn.packages.core.plci_core_data_type import DataType
from muphyn.packages.interface.graphical_actions.diagram_add_input_to_box_action import DiagramAddInputToBoxAction
from muphyn.packages.interface.graphical_actions.diagram_remove_inputs_from_box_action import DiagramRemoveInputsFromBoxAction
from muphyn.packages.interface.models.graphical_models.abstract_box_model import AbstractBoxModel
from muphyn.packages.interface.models.signals_model.input_connection_model import InputConnectionGroupModel
from muphyn.packages.interface.properties_pages.abstract_infinite_io_properties_editor import AbstractInfiniteIOPropertiesEditor
from muphyn.packages.interface.models.graphical_models.box_model import BoxModel
from muphyn.packages.interface.properties_pages.infinite_io_properties_item import InifiniteIOPropertiesItem


#-----------------------------------
# Class
#-----------------------------------

class InfiniteInputPropertiesEditor (AbstractInfiniteIOPropertiesEditor) :
    """Est la classe qui décrit le fonctionnement du panneau pour éditer les propriétés des entrées infinies."""


    # -------------
    # Constructors
    # -------------

    def __init__ (self, box_model : BoxModel, input_group: InputConnectionGroupModel) :

        AbstractInfiniteIOPropertiesEditor.__init__(self, box_model, input_group.name, input_group.data_type, 
            input_group.minimum_count, input_group.maximum_count)
       
        # 
        for input_index, input_ in enumerate(input_group.inputs):
            if input_group.is_infinite : 
                # Create ghost item
                item = QListWidgetItem(self._tbl_connection)

                # Add ghost item to list
                self._tbl_connection.addItem(item)

                # Create io item
                item_widget = InifiniteIOPropertiesItem(input_index+1, input_)

                # Set item geometry
                item.setSizeHint(item_widget.minimumSizeHint())

                # Replace ghost item by IO properties item
                self._tbl_connection.setItemWidget(item, item_widget)

        # Set button enabled if count < maximum count
        self._btn_add.setEnabled(self._tbl_connection.count() < self.maximum_count)

        # Enable remove button if count > mimimum count
        self._btn_remove.setEnabled(self._tbl_connection.count() > self._minimum_count)

        # Set minimum height
        self._tbl_connection.setMinimumHeight(AbstractInfiniteIOPropertiesEditor.ItemHeight)
        self._input_changed_semaphore = False

        # Connect input count changed signal
        self.box_model.input_count_changed.connect(self.box_model_input_count_changed)
        # self.box_model.param_changed.connect(self.on_box_model_param_changed)

        
    # -------------
    # Methods
    # -------------
    
    def contains_input (self, input : Any) -> int : 
        """Permet de savoir si l'entrée est contenue dans la liste."""

        for element_index in range(self._tbl_connection.count()) :
            item = self._tbl_connection.item(element_index)
            item_widget = self._tbl_connection.itemWidget(item)
            item_input = item_widget._connection_model

            if item_input.graphical_index == input.graphical_index :
                return element_index

        return -1


    def box_model_input_count_changed (self) -> None :
        """Est la méthode appelée losque le nombre d'entrée le box est modifié."""

        if self._input_changed_semaphore : 
            return

        # Get inputs group
        inputs_group = self._box_model.get_inputs_group(self._group_name)

        # Remove items
        for element_index in range(self._tbl_connection.count() - 1, -1, -1) :
            item = self._tbl_connection.item(element_index)
            item_widget = self._tbl_connection.itemWidget(item)
            input_ = item_widget._connection_model

            if not(input_ in inputs_group.inputs) :
                item_widget : InifiniteIOPropertiesItem = self._tbl_connection.itemWidget(item)
                self._tbl_connection.takeItem(self._tbl_connection.indexFromItem(item).row())
                item_widget.deleteLater()

        # Adding Items
        for input_index, input_ in enumerate(inputs_group.inputs):
            # Determine if input is already in the list
            element_index = self.contains_input(input_)

            # Add element in the list in the input is already there
            if element_index == -1:
                # Create list ghost item
                item = QListWidgetItem(self._tbl_connection)

                # Add ghost item to the list
                self._tbl_connection.addItem(item)

                # Replace ghost widget by InifiniteIOPropertiesItem
                item_widget = InifiniteIOPropertiesItem(input_index + 1, input_)
                self._tbl_connection.setItemWidget(item, item_widget)

        if self._tbl_connection.count() > 0 :
            item.setSizeHint(self._tbl_connection.itemWidget(self._tbl_connection.item(0)).minimumSizeHint())

        self.recompute_connection_numbers()
        
        # Set button enabled if count < maximum count
        self._btn_add.setEnabled(self._tbl_connection.count() < self.maximum_count)

        # Enable remove button if count > mimimum count
        self._btn_remove.setEnabled(self._tbl_connection.count() > self._minimum_count)

    def translate_ui (self) -> None :
        super().translate_ui()
        
        self._lbl_connection_mode_value.setText(QCoreApplication.translate(self.objectName(), u"Input", None))

    def button_add_click (self) -> None :
        # Build adding input action
        action = DiagramAddInputToBoxAction(
            self.box_model,
            self._group_name
        )
        action.do()
        self.actions_holder.append(action)

    def button_remove_click (self) -> None :
        # Get input to remove
        inputs_group = self.box_model.get_inputs_group(self._group_name)
        inputs_to_remove = [inputs_group.inputs[-1]]
        
        # Build input removal action
        action = DiagramRemoveInputsFromBoxAction(
            self._box_model,
            self._group_name,
            inputs_to_remove
        )

        # Remove input
        action.do()

        # Add action to action holder
        self.actions_holder.append(action)

    def unload (self) -> None :
        # Disconect input count changed event 
        self.box_model.input_count_changed.disconnect(self.box_model_input_count_changed)


#-----------------------------------
# Imports
#-----------------------------------

from typing import List

from muphyn.packages.interface.graphical_actions.abstract_graphical_action import AbstractGraphicalAction

#-----------------------------------
# Class
#-----------------------------------

class ActionsHolder : 
    """Est la classe permettant de lister les actions réalisée."""
    
    # -------------
    # Constructors
    # -------------

    def __init__ (self) :
        self._undo_list : List[AbstractGraphicalAction] = []
        self._redo_list : List[AbstractGraphicalAction] = [] 

    # -------------
    # Properties
    # -------------

    @property
    def undo_count (self) -> int :
        """Permet de récuperer le nombre d'actions contenues dans la liste des actions à défaire."""
        return self._undo_list.__len__()

    @property
    def redo_count (self) -> int :
        """Permet de récuperer le nombre d'actions contenues dans la liste des actions à refaire."""
        return self._redo_list.__len__()

    @property
    def last_action (self) -> AbstractGraphicalAction:
        """Permet de récuperer la dernoère action "done"."""
        
        if len(self._undo_list) == 0 :
            return None

        return self._undo_list[0]

    # -------------
    # Methods
    # -------------

    def clear (self) :
        """Permet de vider les listes d'actions."""
        
        self._undo_list.clear()
        self._redo_list.clear()

    def undo (self) :
        """Permet de réaliser la dernière action contenue dans la liste."""

        if len(self._undo_list) > 0 :

            action = self._undo_list[0]
            action.undo()
            self._redo_list.insert(0, action)
            self._undo_list.remove(action)


    def redo (self) :
        """Permet de re-réaliser la dernière action qui a été dé-faite."""

        if len(self._redo_list) > 0 :

            action = self._redo_list[0]
            action.do()
            self._undo_list.insert(0, action)
            self._redo_list.remove(action)
    
    def append (self, action : AbstractGraphicalAction) : 
        """Permet d'ajouter une action dans la liste."""

        self._undo_list.insert(0, action)
        self._redo_list.clear()

    


#-----------------------------------
# Imports
#-----------------------------------

# PyQt5 Imports
from PyQt5.QtCore import QSizeF

# Project Imports
from muphyn.packages.interface.graphical_actions.abstract_unique_element_diagram_action import AbstractUniqueElementDiagramAction
from muphyn.packages.interface.models.graphical_models.abstract_graphical_element import AbstractGraphicalElement

#-----------------------------------
# Class
#-----------------------------------

class DiagramResizeGraphicalElementAction (AbstractUniqueElementDiagramAction) :
    """Est l'action qui permet de modifier la taille des léments graphiques."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, graphical_element : AbstractGraphicalElement, old_size : QSizeF, new_size : QSizeF) :
        
        AbstractUniqueElementDiagramAction.__init__(self, graphical_element)

        self._old_size = old_size
        self._new_size = new_size

    # -------------
    # Properties
    # -------------

    @property
    def old_size (self) -> QSizeF :
        """Permet de récuperer la taille avant la modification."""
        return self._old_size

    @property
    def new_size (self) -> QSizeF :
        """Permet de récuperer la taille après modification."""
        return self._new_size

    @new_size.setter
    def new_size (self, new_size_ : QSizeF) -> None :
        """Permet de modifier la taille qui doit être appliqué pour la modification."""
        self._new_size = new_size_
        
    # -------------
    # Methods
    # -------------

    def do (self) :
        
        graphical_element = self.graphical_element

        if graphical_element is None : 
            return

        graphical_element.action_size_semaphore = True
        graphical_element.size = self.new_size
        graphical_element.action_size_semaphore = False

    def undo (self) :
        
        graphical_element = self.graphical_element

        if graphical_element is None : 
            return

        graphical_element.action_size_semaphore = True
        graphical_element.size = self.old_size
        graphical_element.action_size_semaphore = False


#-----------------------------------
# Imports
#-----------------------------------

from typing import Any

from PyQt5.QtCore import QPointF, QRectF, QSizeF, Qt, pyqtSignal
from PyQt5.QtGui import QPainter
from PyQt5.QtWidgets import QGraphicsItem, QGraphicsObject
from numpy import minimum

from muphyn.packages.interface.models.event_signal_data import EventSignalData

#-----------------------------------
# Class
#-----------------------------------

class AbstractResizer (QGraphicsObject) :
    """Est la classe abstraite commune des éléments capable de redimensionner les boxes."""

    # -------------
    # Signals
    # -------------

    resizeSignal = pyqtSignal(EventSignalData)
    selectedSignal = pyqtSignal(EventSignalData)

    # -------------
    # Constructors
    # -------------

    def __init__ (self, parent, size : QSizeF) :

        QGraphicsObject.__init__(self, parent)

        self.setFlag(QGraphicsItem.ItemIsMovable, True)
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges, True)

        self.setZValue(5)

        self._rect = QRectF(0, 0, size.width(), size.height())
        self._parent = parent

    # -------------
    # Methods
    # -------------

    def boundingRect (self) :
        return self._rect

    def paint (self, painter : QPainter, option, widget = None) :

        if self.isSelected() :
            painter.drawRect(self._rect)

        else :
            painter.fillRect(self._rect, Qt.black)

    def itemChange (self, change: QGraphicsItem.GraphicsItemChange, value: Any) -> Any :
        
        if change == QGraphicsItem.ItemSelectedChange :
            self.selectedSignal.emit(EventSignalData(self, value))

        elif change == QGraphicsItem.ItemPositionChange :
            if self.isSelected() :
                
                value : QPointF = self.changeValue(value)

                if not(self._parent == None) : 

                    new_width = self._parent._bound.width() + value.x()
                    new_height = self._parent._bound.height() + value.y()

                    minimum_size = self._parent.minimum_size

                    if new_width < minimum_size.width() :
                        value.setX(minimum_size.width() - self._parent._bound.width())

                    if new_height < minimum_size.height() :
                        value.setY(minimum_size.height() - self._parent._bound.height())
                    

                if not(value.x() == 0 and value.y() == 0) :
                    self.resizeSignal.emit(EventSignalData(self, value))
            
                return QPointF(self.pos().x() + (value.x() / 2), self.pos().y() + (value.y() / 2))

        return super().itemChange(change, value)

    def changeValue (self, value_ : QPointF) -> QPointF :
        raise Exception('AbstractResizer.changeValue is an abstract method and must be overhidden.')


#-----------------------------------
# Imports
#-----------------------------------

from typing import List, Any, Dict
from os import listdir, path
from threading import Lock

from PyQt5.QtCore import QFileInfo

from muphyn.packages.core.plci_core_data_type import DataType
from muphyn.packages.core.plci_core_diagram import Diagram
from muphyn.packages.core.plci_core_box import Box

# from muphyn.packages..core.box_library.box_library_importer_1_0_0 import BoxLibraryImporter_1_0_0
from muphyn.packages.core.box_library.box_library_importer_1_1_0 import BoxLibraryImporter_1_1_0
from muphyn.packages.core.box_library.abstract_box_library_importer import AbstractBoxLibraryImporter
from muphyn.packages.core.box_library.box_library_data import AbstractBoxData
from muphyn.packages.core.utils.log_manager import LogManager
from muphyn.packages.core.utils.singleton import SingletonMetaClass

#-----------------------------------
# Classes
#-----------------------------------

class BoxLibraryElement : 
    """Est la classe qui permet de stocker les éléments de bibliothèque dans la classe boxes."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, path : str) :
        self._path = path
        self._loaded = False
        self._boxes : Dict[str, AbstractBoxData] = {}

    # -------------
    # Properties
    # -------------
    
    @property
    def path (self) -> str : 
        """Permet de récuperer le lien vers le dossier de la bibliothèque de boxes.""" 
        return self._path

    @property
    def loaded (self) -> bool :
        """Permet de récuperer l'état de chargement de la bibliothèque de boxes."""
        return self._loaded

    @property
    def boxes (self) -> Dict[str, AbstractBoxData] :
        """Permet de récuperer le dictionnaire des nom de bibliothèque et leur éléments de création de boxes."""
        return self._boxes

    # -------------
    # Methods
    # -------------

    def load (self, box_importer : AbstractBoxLibraryImporter) :
        """Permet de charger la bibliothèque."""
        # Récuperation des fichiers dans le dossier
        for current_file in listdir(self.path) :

            if current_file.endswith('.yaml') :
                file_name = QFileInfo(current_file).completeBaseName()
                absolute_yaml_file = self.path + '/' + file_name + '.yaml'
                
                imported_box_data = box_importer.import_box(self.path, file_name, absolute_yaml_file, self._boxes)

                if imported_box_data is None : 
                    continue

                if imported_box_data['box_data'] is None :
                    continue

                self._boxes[imported_box_data['library_name']] = imported_box_data['box_data']

class BoxesLibrariesManager(metaclass=SingletonMetaClass) :
    """Est la classe qui permet de construire les boxes.""" 

    # -------------
    # Constructors
    # -------------
    def __init__ (self) :
        self._libraries : List[BoxLibraryElement] = []
        self._current_box_index = 0
        self._box_importer = BoxLibraryImporter_1_1_0()
    
    # -------------
    # Properties
    # -------------

    @property
    def current_box_index (self) -> int :
        """Permet de retourner l'index actuelle de la création des boxes."""
        return self._current_box_index

    @property
    def libraries (self) -> List[BoxLibraryElement] :
        """Permet de retourner la liste des libraries."""
        return self._libraries

    @property 
    def box_importer (self) -> AbstractBoxLibraryImporter :
        """Permet de retourner l'importeur utilisé pour importer des boxes."""
        return self._box_importer

    @property
    def boxes (self) -> List[AbstractBoxData] :
        """Permet de retourner l'intégralité des boxes importées."""

        for library in self._libraries :
            for box_name in library.boxes :
                yield library.boxes[box_name]


    # -------------
    # Methods
    # -------------

    def add_library (self, library_folder : str) :
        """Permet d'ajouter une bibliothèque dans le répertoire des bibliothèque."""

        
        # Test if library_folder path format is string 
        if not (isinstance(library_folder, str)) :
            LogManager().error(f"Wrong path variable format {type(library_folder)} instead of str", is_global_message=True)
            return False

        # Test if library folder path exists
        if not path.exists(library_folder):
            LogManager().error(f"Wrong path variable format {type(library_folder)} instead of str", is_global_message=True)
            return False

        # Check if library has already been added
        if any(libraryElement.path == library_folder for libraryElement in self._libraries):
            LogManager().error(f"Wrong path variable format {type(library_folder)} instead of str", is_global_message=True)
            return False

        # Append Scheduler Library
        self._libraries.append(BoxLibraryElement(library_folder))

        return True

    def load_libraries (self) :
        """Permet de charger toutes les bibliothèques du répertoire."""

        for libraryElement in self._libraries :

            if not libraryElement.loaded :
                libraryElement.load(self.box_importer)

    def construct_box (self, box_library : str, box_name : str, **box_params) -> Any:
        """Permet de construire une box suivant son nom et sa librairie."""

        if not (isinstance(box_library, str) and isinstance(box_name, str)) :
            return None

        box_access = self._name_library(box_library, box_name)

        for libraryElement in self._libraries :
            
            if box_access in libraryElement.boxes :
                box = libraryElement.boxes[box_access].construct_box(self._current_box_index, box_params, self)
                
                if isinstance(box, Diagram) :
                    self._current_box_index = box._boxes[box._boxes.__len__() - 1].index + 1

                elif isinstance(box, Box) :
                    self._current_box_index += 1

                return box

    def get_required_params (self, box_library : str, box_name : str) -> Dict[str, DataType] :
        """Permet de retourner les paramètres nécessaires pour instancier une box dans une bibliothèque."""

        if not (isinstance(box_library, str) and isinstance(box_name, str)) :
            return None

        box_access = self._name_library(box_library, box_name)

        for libraryElement in self._libraries :
            
            if box_access in libraryElement.boxes :
                boxElement = libraryElement.boxes[box_access]
                return boxElement.params

        return None
        
    def _name_library (self, box_library : str, box_name : str) -> str :
        """Permet de rassembler le nom et la libraire en un seul string."""
        return box_library + "." + box_name

    def clear (self) -> None :
        """Permet d'éffacer le contenu des boxes chargées."""
        for libraryElement in self._libraries :

            libraryElement.boxes.clear()
            del libraryElement
        
        del self._libraries
        self._libraries : List[BoxLibraryElement] = []

    def get_box_data (self, box_library : str, box_name : str) -> AbstractBoxData : 
        """Permet de récuperer les données de construction d'une box en fonction de sa bibliothèque et de son nom."""

        for box_data in self.boxes :

            if box_data.box_library == box_library and box_data.box_name == box_name :
                return box_data

        return None


# General import
from typing import Optional, Union

# PyQt import
from PyQt5.QtWidgets import QWidget

# Project import
from muphyn.packages.interface.widgets.DrawingWidget import DrawingWidget
from muphyn.packages.interface.editors.abstract_editor import AbstractEditor
from muphyn.packages.interface.editors.multiphysics_model import AbstractComponentModel, ConnectionModel
from muphyn.packages.core.utils.log_manager import LogManager

class AbstractMultiPhysicsModel(DrawingWidget):

    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)

        self.components: list[AbstractComponentModel] = []
        self.links: list[ConnectionModel] = []

    def addItem(self, newItem: Union[AbstractComponentModel, ConnectionModel]):
        if isinstance(newItem, AbstractComponentModel):
            # Test if already added to avoid duplicated item in the scene
            if newItem not in self.components:
                # Add item to list
                self.components.append(newItem)

                # Add item to scene
                self.scene.addItem(newItem)
            else:
                LogManager.debug("Component already added")

        elif isinstance(newItem, ConnectionModel):
            # Test if already added to avoid duplicated item in the scene
            if newItem not in self.links:
                # Add item to list
                self.links.append(newItem)

                # Add item to scene
                self.scene.addItem(newItem)
            else:
                LogManager.debug("Connection already added")
        else:
            raise(TypeError(f"AbstractMultiPhysicsModel.addItems(): newItem is not a AbstractComponentModel → {type(newItem)}"))

        
        


#-----------------------------------
# Imports
#-----------------------------------

from typing import List

#-----------------------------------
# Functions
#-----------------------------------

def _init_derivator_box (box, simulation_params) -> None :
    box['last_output'] = 0
    box['last_input'] = 0
    box['points_evaluated'] = 0

def _function_derivator_box (box, event_) -> List :

    v = 0
    input = box.get_input(0).value

    if (not event_.step_time == 0) and (not event_.timing == 0) : 
        v = (input - box['last_input']) / event_.step_time
        
    else : 
        return []

    box['last_output'] = v
    box['last_input'] = input
    box['points_evaluated'] = box['points_evaluated'] + 1

    if box['points_evaluated'] < 2 :
        return []

    events : List = []
    
    for output in box.outputs :
        events.append(box.construct_signal_event(output, v))

    return events



#
from muphyn.packages.core.plci_core_data_type import DataType
from muphyn.packages.core.utils.log_manager import LogManager
from muphyn.packages.interface.properties_pages.abstract_properties_editor import AbstractPropertiesEditor

# DataType {FLOAT, INT, BOOLEAN, STRING}
from muphyn.packages.interface.properties_pages.data_property_widgets.basic_type_property_widgets import \
    BooleanPropertyWidget, DoublePropertyWidget, IntegerPropertyWidget, StringPropertyWidget, UnknownTypePropertyWidget, PropertyLineEdit

# Datatype {ANYFILE, DIRECTORY, EXISTINGFILE, EXISTINGFILES}
from muphyn.packages.interface.properties_pages.data_property_widgets.path_property_widgets import \
    AnyFilePropertyWidget, DirectoryPropertyWidget, ExistingFilePropertyWidget, ExistingFilesPropertyWidget

from muphyn.packages.interface.properties_pages.data_property_widgets.choice_property_widget import ChoicePropertyWidget

def property_widget_factory(parameter_to_edit: dict) -> AbstractPropertiesEditor:
    # Get parameter type name
    param_type_name = parameter_to_edit["type"].__str__().lower()


    if param_type_name == str(DataType.BOOLEAN):
        return BooleanPropertyWidget()

    elif param_type_name == str(DataType.FLOAT):
        return PropertyLineEdit(parameter_to_edit, DataType.FLOAT)
        # return DoublePropertyWidget(parameter_to_edit)

    elif param_type_name == str(DataType.INT):
        return PropertyLineEdit(parameter_to_edit, DataType.INT)
        # return IntegerPropertyWidget(parameter_to_edit)

    elif param_type_name == str(DataType.STRING):
        # return  PropertyLineEdit(parameter_to_edit, DataType.STRING)
        return  StringPropertyWidget(parameter_to_edit)

    elif param_type_name == str(DataType.ANYFILE):
        return AnyFilePropertyWidget()

    elif param_type_name == str(DataType.DIRECTORY):
        return DirectoryPropertyWidget()

    elif param_type_name == str(DataType.EXISTINGFILE):
        return ExistingFilePropertyWidget()

    elif param_type_name == str(DataType.EXISTINGFILES):
        return ExistingFilesPropertyWidget()

    elif param_type_name == str(DataType.CHOICE):
        return ChoicePropertyWidget(parameter_to_edit)
    else:
        LogManager().error(f"Unsupported parameter type for : {param_type_name}")
        return UnknownTypePropertyWidget(param_type_name.__str__())




#-----------------------------------
# Imports
#-----------------------------------

from typing import Callable, List, Any, Dict

from muphyn.utils.paths import ROOT_DIR

from ..plci_core_box import Box
from ..plci_core_data import Data
from ..plci_core_data_type import DataType, get_data_type
from ..plci_core_diagram import Diagram
from ..plci_core_signal import Signal
from ..plci_core_signal_event import SignalEvent
from ..plci_core_scheduler_event import SchedulerEvent

#-----------------------------------
# Methods
#-----------------------------------

def construct_diagram (signals : Dict, boxes : Dict, inputs : Dict, outputs : Dict, box_params : Dict[str, Any], boxes_library : Any) -> Diagram :
   
    diagram = Diagram()
    
    for signal_data in signals :
        
        data_type : DataType = get_data_type(signal_data['type'])
        signal : Signal = Signal(signal_data['index'], data_type, signal_data['value']) 
        diagram.append(signal)

    for box_data in boxes :

        params = box_data['params']
        if params == 'None' :
            box = boxes_library.construct_box(box_data['library'], box_data['name'])
        else :
            box = boxes_library.construct_box(box_data['library'], box_data['name'], **params)
        
        if isinstance(box, Diagram) :
            diagram.append(box)

            if not box_data['inputs'] == 'None' :

                for signal_index, input in enumerate(box_data['inputs']) :
                    signal = diagram.signals[input['signal_index']]
                    
                    input_signal = box.signals[box.inputs[signal_index]]
                    diagram.add_linked_signals(signal, input_signal)

            if not box_data['outputs'] == 'None' :

                for signal_index, output in enumerate(box_data['outputs']) :
                    signal = diagram.signals[output['signal_index']]
                    
                    output_signal = box.signals[box.outputs[signal_index]]
                    diagram.add_linked_signals(output_signal, signal)

        elif isinstance(box, Box) :
            diagram.append(box)

            if not box_data['inputs'] == 'None' :

                for box_input_data in box_data['inputs'] :
                    diagram.add_box_inputs(box, diagram.signals[box_input_data['signal_index']])

            if not box_data['outputs'] == 'None' :
                
                for box_output_data in box_data['outputs'] :
                    diagram.add_box_outputs(box, diagram.signals[box_output_data['signal_index']])

    for input in inputs :
        diagram.inputs.append(input['signal_index'])

    for output in outputs :
        diagram.outputs.append(output['signal_index'])

    return diagram


#-----------------------------------
# Classes
#-----------------------------------

class AbstractBoxData : 
    """Est la classe commune des classes de données de bibliothèque."""

    # -------------
    # Constructors
    # -------------

    def __init__ (
            self, 
            box_name: str, 
            box_library: str, 
            box_type: str,
            creator: str,
            date_created: Any,
            version: float,
            inputs,
            outputs,
            icon: str = None
        ) :
        self._box_name = box_name
        self._box_library = box_library
        self._box_type = box_type
        self._creator : str = creator
        self._date_created : Any = date_created
        self._version : float = version
        self._inputs: dict[str, dict] = inputs
        self._outputs: dict[str, dict] = outputs
        self._full_box_path = self._box_library + '.' + self._box_name
        if icon is not None:
            self._icon = ROOT_DIR + "/" + icon
        else:
            self._icon = None

    # -------------
    # Properties
    # -------------

    @property 
    def box_name (self) -> str :
        """Permet de récuperer le nom de la box."""
        return self._box_name

    @property
    def box_library (self) -> str :
        """Permet de récuperer la bibliothèque dans laquel se trouve la box."""
        return self._box_library

    @property
    def box_type (self) -> str :
        """Permet de récuperer la bibliothèque dans laquel se trouve la box."""
        return self._box_type

    @property
    def creator (self) -> str :
        """Permet de récuperer le nom de la personne qui a créé la box."""
        return self._creator

    @property
    def date_created (self) -> Any :
        """Permet de récuperer la date à laquelle la box a été créée."""
        return self._date_created

    @property
    def version (self) -> float :
        """Permet de récuperer la version de la box."""
        return self._version

    @property
    def inputs (self) :
        """Permet de récuperer la liste des signaux considérés comme entrées."""
        return self._inputs

    @property
    def outputs (self) :
        """Permet de récuperer la liste des signaux considérés comme sorties."""
        return self._outputs

    @property
    def full_box_path (self) -> str :
        """Permet de récuperer le nom complet de la boite (bibliothèque + nom)."""
        return self._full_box_path

    @property
    def icon(self) -> str:
        return self._icon

    # -------------
    # Methods
    # -------------

    def __lt__ (self, other) :
        return self._full_box_path.__lt__(other._full_box_path) 


    def construct_box (self, index, box_params : Dict[str, Any], boxes_library : Any) -> Any :
        """Permet de générer la box."""
        raise Exception('AbstractBoxData construct_box is an abstract method and must be overriden !')

class CompositeBoxData (AbstractBoxData) :
    """Est la classe qui permet de contenir les données des boxes composites."""
    
    # -------------
    # Constructors
    # -------------

    def __init__ (
            self, 
            box_name : str, 
            box_library : str, 
            creator : str,
            date_created : Any,
            version : float,
            signals, 
            boxes, 
            inputs, 
            outputs,
            icon: str = None
        ) :
        AbstractBoxData.__init__(self, box_name, box_library, creator, date_created, version, inputs, outputs, icon)
        self._signals = signals
        self._boxes = boxes
    
    # -------------
    # Properties
    # -------------
    
    def signals (self) :
        """Permet de récuperer le dictionnaire contenant la composition des signaux de la box composite."""
        return self._signals

    def boxes (self) :
        """Permet de récuperer le dictionnaire contenant la composition des boxes de la box composite."""
        return self._boxes

    # -------------
    # Methods
    # -------------

    def construct_box (self, index, box_params : Dict[str, Any], boxes_library : Any) -> Diagram :
        """Permet de générer la box."""
        
        return construct_diagram(self._signals, self._boxes, self._inputs, self._outputs, box_params, boxes_library)

class CodeBoxData (AbstractBoxData) :
    """Est la classe qui permet de contenir les données des boxes importées par un bibliothèque."""

    # -------------
    # Constructors
    # -------------
    def __init__ (
            self, 
            box_name : str, 
            box_library : str, 
            box_type: str,
            wait_for_event : bool, 
            wait_for_all_signal_events : bool,
            params : Dict[str, Data], 
            box_init_method_ : Callable[[Box], None], 
            box_function_ : Callable[[Box, SchedulerEvent], List[SignalEvent]], 
            box_end_method_ : Callable[[Box], None],
            creator : str,
            date_created : Any,
            version : float, 
            inputs,
            outputs,
            icon: str = None
        ) :

        AbstractBoxData.__init__(self, box_name, box_library, box_type, creator, date_created, version, inputs, outputs, icon)
        self._init_method : Callable[[Box], None] = box_init_method_
        self._box_function : Callable[[Box, SchedulerEvent], List[SignalEvent]] = box_function_
        self._end_mehtod : Callable[[Box], None] = box_end_method_
        self._params : Dict[str, Data] = params
        self._wait_for_event = wait_for_event
        self._wait_for_all_signal_events = wait_for_all_signal_events

    # -------------
    # Properties
    # -------------

    @property
    def init_method (self) -> Callable[[Box], None] :
        """Permet de récuperer la méthode d'initialisation de la box."""
        return self._init_method
    
    @property
    def box_function (self) -> Callable[[Box, SchedulerEvent], List[SignalEvent]] :
        """Permet de récuperer la fonction de la box."""
        return self._box_function

    @property
    def end_method (self) -> Callable[[Box], None] :
        """Permet de récuperer la méthode de fin de la box."""
        return self._end_mehtod

    @property
    def params (self) -> Dict[str, Data] : 
        """Permet de récuperer tous les paramètres requis pour la génération de ce type de box."""
        return self._params

    @property 
    def wait_for_event (self) -> bool :
        """Permet de récuperer le fait qu'il faille attendre un événement pour que la box puisse s'activer."""
        return self._wait_for_event

    @property
    def wait_for_all_signal_events (self) -> bool :
        """Permet de récuperer le fait qu'il faille que toutes les entrées de la box aie eu des événements au timing actuel pour qu'elle s'active."""
        return self._wait_for_all_signal_events

    # -------------
    # Methods
    # -------------
    
    def construct_box (self, index, box_params : Dict[str, Any], boxes_library : Any) -> Box :
        """Permet de générer la box."""
        return Box(
                    index_ = index,
                    name_ = self.box_name,
                    library_ = self.box_library,
                    init_method_ = self.init_method,
                    function_ = self.box_function,
                    end_method_ = self.end_method,
                    wait_for_events_ = self.wait_for_event,
                    wait_for_all_signal_events_ = self.wait_for_all_signal_events,
                    params_ = box_params
                )

class MultiPhysicsSimulationBoxData(CodeBoxData):
    def __init__(self, box_name: str, box_library: str, box_type: str, wait_for_event: bool, wait_for_all_signal_events: bool, params: Dict[str, Data], box_init_method_: Callable[[Box], None], box_function_: Callable[[Box, SchedulerEvent], List[SignalEvent]], box_end_method_: Callable[[Box], None], creator: str, date_created: Any, version: float, inputs, outputs, icon: str = None):
        super().__init__(box_name, box_library, box_type, wait_for_event, wait_for_all_signal_events, params, box_init_method_, box_function_, box_end_method_, creator, date_created, version, inputs, outputs, icon)

    @staticmethod
    def fromCodeBoxData(code_box_data: CodeBoxData):

        return MultiPhysicsSimulationBoxData(
            code_box_data.box_name,
            code_box_data.box_library,
            code_box_data.box_type,
            code_box_data.wait_for_event,
            code_box_data.wait_for_all_signal_events,
            code_box_data.params,
            None if not hasattr(code_box_data, "box_init_method_") else code_box_data.box_init_method_, 
            None if not hasattr(code_box_data, "box_function_") else code_box_data.box_function_, 
            None if not hasattr(code_box_data, "box_end_method_") else code_box_data.box_end_method_,
            code_box_data.creator,
            code_box_data.date_created,
            code_box_data.version,
            code_box_data.inputs,
            code_box_data.outputs,
            code_box_data.icon
        )



from PyQt5.QtCore import QPointF, QSizeF, Qt
from PyQt5.QtWidgets import QGraphicsItem
from PyQt5.QtGui import QColor

from basic_shape.shapes import Text
from basic_shape.image_shapes import ImageShapeFactory
from basic_shape.grouped_shapes import GroupedShapes
from electric_components.electric_scheme_part.electrical_connector import ElectricalConnector

class Resistor(GroupedShapes):

    # Image 
    border_color = QColor(0, 0, 0)

    background_color = QColor(255, 255, 255)

    rectangle_size = QSizeF(100, 100)

    def __init__(self, group_position: QPointF = QPointF(0, 0), parent: QGraphicsItem = None) -> None:

        super().__init__(group_position, parent=parent)

        # Init Positive Node
        self.positive_node = ElectricalConnector(
            ElectricalConnector.ElectricalConnectorType.Positive,
            position=QPointF(0, Resistor.rectangle_size.height()/2),
            parent=self
        )
        
        # Init Negative Node
        self.negative_node = ElectricalConnector(
            ElectricalConnector.ElectricalConnectorType.Negative,
            QPointF(Resistor.rectangle_size.width() + ElectricalConnector.connector_size.width(), Resistor.rectangle_size.height()/2),
            parent=self
        )
        self.negative_node.setRotation(180)
        
        # Init Resistor Icon
        self.resistor_icon = ImageShapeFactory(
            "packages/interface/assets/ec-Resistor_US.svg", 
            QPointF(ElectricalConnector.connector_size.width(), 0), Resistor.rectangle_size, 
            border_color=QColor(255, 0, 0),
            aspect_ratio_mode=Qt.AspectRatioMode.KeepAspectRatio,
            parent=self
        )
        self.resistor_icon.setRotation(90)

        # Text 
        self.text = Text(
            "R", 
            QPointF(60, 10), 60,
            parent=self)

        self.setFlag(QGraphicsItem.ItemIsMovable)


from PyQt5.QtCore import Qt
from PyQt5.QtGui import QColor, QFont, QPen

MuphynColor = QColor(233, 125, 26)
ApplicationWindowTitle = "Muphyn"

# Text Font
class MuphynFonts:
    """
    This class saves all Fonts data for Muphyn application
    """
    MuphynFontFamily: str = "MS Shell Dlg 2"

    # Input / Output Connector Font
    IOConnectorFont: QFont = QFont(MuphynFontFamily, 10, weight=QFont.Weight.Normal, italic=False)

    # Box Model Details Font
    BoxModelDetailsFont: QFont = QFont(MuphynFontFamily, 10, weight=QFont.Weight.Bold, italic=False)

class MuphynPens:
    """
    This class saves all Pens data for Muphyn application
    """

    # Link Pens
    SelectedLinkPen = QPen(Qt.GlobalColor.black, 2, Qt.PenStyle.DashLine, Qt.PenCapStyle.RoundCap, Qt.PenJoinStyle.RoundJoin)
    UnSelectedLinkPen = QPen(Qt.GlobalColor.black, 1, Qt.PenStyle.SolidLine, Qt.PenCapStyle.RoundCap, Qt.PenJoinStyle.RoundJoin)



#-----------------------------------
# Imports
#-----------------------------------

from typing import List, Any

from PyQt5.QtWidgets import QWidget, QVBoxLayout, QLayout

from muphyn.packages.interface.graphical_actions.actions_holder import ActionsHolder
from muphyn.packages.interface.properties_pages.abstract_properties_editor import AbstractPropertiesEditor
from muphyn.packages.interface.properties_pages.properties_page_builder import get_properties_page

#-----------------------------------
# Class
#-----------------------------------

class PropertiesWidget (QWidget) :
    """Est le widget qui affiche les propriétés des éléments sélectionnés."""
        
    # -------------
    # Constructors
    # -------------

    def __init__(self, parent : QWidget = None) :
        
        QWidget.__init__(self, parent)

        # Init VBox Layout
        vBoxLayout = QVBoxLayout()
        vBoxLayout.setSizeConstraint(QLayout.SizeConstraint.SetMinAndMaxSize)
        self.setLayout(vBoxLayout)

        self._properties_editors : List[AbstractPropertiesEditor] = []
        self._actions_holder = None
        self._current_model = None
        
    # -------------
    # Properties
    # -------------

    @property
    def actions_holder (self) -> ActionsHolder :
        """Permet de récuperer l'actions holder à utiliser pour les widgets de propriétés."""
        return self._actions_holder

    @actions_holder.setter
    def actions_holder (self, actions_holder_ : ActionsHolder) -> None :
        """Permet de modifier l'actions holder à utiliser pour les widgets de propriétés."""

        self._actions_holder = actions_holder_

        for property_editor in self._properties_editors :
            property_editor.actions_holder = self._actions_holder

    @property
    def current_model (self) -> Any :
        """Permet de récuperer le modèle actuellement en cours d'édition."""
        return self._current_model

    @current_model.setter
    def current_model (self, current_model_ : Any) -> None :
        """Permet de modifier le modèle actuellement en cours d'édition."""
        for editor in self._properties_editors : 

            # Remove all items from layout
            if self.layout() is not None:
                self.layout().removeWidget(editor)

            # Disable update from property editor
            editor.setUpdatesEnabled(False)
            editor.actions_holder = None
            editor.unload()
            editor.deleteLater()

        # Clear list of property editors
        self._properties_editors.clear()

        self._current_model = current_model_

        # Build List of Property Editors
        for rowIndex, editor in enumerate(get_properties_page(self._current_model)):
            editor.actions_holder = self.actions_holder
            self._properties_editors.append(editor)
            self.layout().addWidget(editor)


#-----------------------------------
# Imports
#-----------------------------------

# Project Imports
from muphyn.packages.interface.graphical_actions.abstract_unique_element_diagram_action import AbstractUniqueElementDiagramAction
from muphyn.packages.interface.models.graphical_models.abstract_graphical_element import AbstractGraphicalElement

#-----------------------------------
# Class
#-----------------------------------

class DiagramAddInputToBoxAction (AbstractUniqueElementDiagramAction) :
    """Est la classe d'action permettant d'ajouter une entrée à une box."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, graphical_element : AbstractGraphicalElement, input_group_name : str, input_index: int = -1) :

        AbstractUniqueElementDiagramAction.__init__(self, graphical_element)

        self._input_added_graphical_index = None
        self._input_group_name = input_group_name
        self._input_index = input_index

    # -------------
    # Methods
    # -------------

    def do (self) :
        
        # Append input
        if self._input_index == -1:
            input_model = self.graphical_element.append_input(self._input_group_name)
        else:
            input_model = self.graphical_element.insert_input(self._input_group_name, self._input_index)
        
        if self._input_added_graphical_index is None :
            # Get the input graphical index if not already added before
            self._input_added_graphical_index = input_model.graphical_index

        else :
            # Set the input graphical index if already added before
            input_model.graphical_index = self._input_added_graphical_index


    def undo (self) :
        
        box = self.graphical_element

        if box is None : 
            return

        for input_index in range(box.input_len - 1, -1, -1) :

            input = box.get_input(self._input_group_name, input_index)
            if input.graphical_index == self._input_added_graphical_index :             
                box.remove_input(self._input_group_name, input)


#-----------------------------------
# Imports
#-----------------------------------

from typing import List, Dict

from muphyn.packages.interface.graphical_actions.abstract_unique_element_diagram_action import AbstractUniqueElementDiagramAction
from muphyn.packages.interface.models.graphical_models.abstract_box_model import AbstractBoxModel
from muphyn.packages.interface.models.signals_model.input_connection_model import InputConnectionModel
from muphyn.packages.interface.graphical_actions import parser_decode as parser_decode
from muphyn.packages.interface.graphical_actions import parser_encode as parser_encode

#-----------------------------------
# Class
#-----------------------------------

class DiagramRemoveInputsFromBoxAction (AbstractUniqueElementDiagramAction) :
    """Est l'action permettant de supprimer les entrées d'une box."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, box_element : AbstractBoxModel, input_group_name: str, inputs : list[InputConnectionModel]) :

        AbstractUniqueElementDiagramAction.__init__(self, box_element)

        self._reconstructors : List[Dict] = []
        self._input_group_name: str = input_group_name
        self._inputs: list[InputConnectionModel] = inputs


    # -------------
    # Methods
    # -------------
    def do (self):

        box_model: AbstractBoxModel = self.graphical_element

        if box_model is None : 
            return

        for input_ in self._inputs:
            input_dict : Dict = {"graphical_index": input_.graphical_index}
            current_input : Dict = parser_encode.box_input(input_, input_dict)

            for signal in input_.links : 
                current_input['signals'].append(parser_encode.link(signal))
                signal.unbind()
                self.diagram_model.remove_element(signal)

            input_._links.clear()
            
            box_model.remove_input(self._input_group_name, input_)

            self._reconstructors.append(input_dict)

 
    def undo (self) :
        
        box_model : AbstractBoxModel = self.graphical_element 

        if box_model is None : 
            return

        self._reconstructors.reverse()
        for input_dict in self._reconstructors :
            parser_decode.box_input(input_dict, box_model)

            for signal_dict in input_dict['signals'] : 
                parser_decode.link(signal_dict, self.diagram_model)


#-----------------------------------
# Imports
#-----------------------------------

# General Imports
from operator import itemgetter
from typing import Iterable, overload, Union

# PyQt5 Imports
from PyQt5.QtCore import QLineF, QPointF, Qt, QRectF
from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import QGraphicsItem, QGraphicsSceneHoverEvent, QGraphicsSceneMouseEvent

# Project Imports
from muphyn.packages.core.plci_core_data import Data
from muphyn.packages.core.plci_core_data_type import DataType
from muphyn.packages.core.utils.log_manager import LogManager
from muphyn.packages.interface.models.graphical_models.abstract_graphical_element import AbstractGraphicalElement
from muphyn.packages.interface.models.signals_model.abstract_connection_model import AbstractConnectionModel
from muphyn.packages.interface.models.signals_model.signal_link_model import SignalLinkModel
from muphyn.packages.interface.models.signals_model.signal_node_model import start_drag
from muphyn.packages.interface.shapes.basic_shape.grouped_shapes import GroupedShapes
from muphyn.packages.interface.shapes.basic_shape.shapes import Line, Text, Polygon
from muphyn.packages.interface.utils.interface_constants import MuphynFonts


#-----------------------------------
# Class
#-----------------------------------

class OutputConnectionModel (AbstractConnectionModel) :
    """Est la classe décrivant le fonctionnement des sorties des boxes."""

    # -------------
    # Static Values
    # -------------
    ConnectorPoints = [
        QPointF(AbstractConnectionModel.ConnectorLineLength, AbstractConnectionModel.ConnectorSide - 3),
        QPointF(AbstractConnectionModel.ConnectorLineLength, 2*AbstractConnectionModel.ConnectorSide - 3),
        QPointF(AbstractConnectionModel.ConnectorLineLength + AbstractConnectionModel.ConnectorSide, 3*AbstractConnectionModel.ConnectorSide/2 - 3)
    ]
    LinePoints = [
        QPointF(0, 3*AbstractConnectionModel.ConnectorSide/2 - 3),
        QPointF(AbstractConnectionModel.ConnectorLineLength, 3*AbstractConnectionModel.ConnectorSide/2 - 3)
    ]

    # -------------
    # Constructors
    # -------------

    def __init__ (self, name : str, data_type : DataType, position : QPointF, link : SignalLinkModel = None,
                  default_value : Data = None, text : str = '', is_infinite : bool = False, parent : QGraphicsItem = None) :
        
        links : Iterable[SignalLinkModel] = []
        if link is not None:
            links.append(link)

        AbstractConnectionModel.__init__(self, name, data_type, position, AbstractConnectionModel.ItemSize, links, text, is_infinite, parent)

        self._default_value : Data = default_value

        # Get constants
        side = AbstractConnectionModel.ConnectorSide
        line_length = AbstractConnectionModel.ConnectorLineLength

        # Connector Group
        self.connector_group = GroupedShapes(parent=self)
        self.connector_group.setAcceptHoverEvents(True)
        self.connector_group.hoverEnterEvent = self.connectorHoverEnterEvent
        self.connector_group.hoverLeaveEvent = self.connectorHoverLeaveEvent

        # Circle
        self._connector: Polygon = Polygon( 
            OutputConnectionModel.ConnectorPoints, 
            border_color=self._color,
            fill_color=Qt.GlobalColor.white,
            pen_join_style=Qt.PenJoinStyle.RoundJoin, 
            parent=self.connector_group
        )
        self.connector_center = self._connector.boundingRect().center()

        # Line
        start_point = QPointF(0, 3*side/2 - 3)
        end_point = QPointF(line_length, 3*side/2 - 3) 
        self._line = Line(QLineF(*OutputConnectionModel.LinePoints), self._color, parent=self.connector_group)

        # Text
        self._lbl: Text = Text(
            name, QPointF(AbstractConnectionModel.ItemWidth+3, 0), 
            alignment=Qt.AlignmentFlag.AlignBottom | Qt.AlignmentFlag.AlignLeft,
            font=MuphynFonts.IOConnectorFont, parent=self
        )

        # Delete label bounding rectangle
        # So it doesn't increase the connector & BoxModel bounding rectangle with the size of the text
        # self._lbl.boundingRect = lambda: QRectF()

        # Hide label
        self._lbl.hide()

    # -------------
    # Properties
    # -------------
    @property
    def default_value (self) -> Data :
        """Permet de récuperer la valeur par défaut du signal."""
        return self._default_value

    @default_value.setter
    def default_value (self, default_value_ : Data) -> None :
        """Permet de modifier la valeur par défaut du signal."""
        self._default_value = default_value_

    @property
    def is_connected_to_output (self) -> bool :
        """Permet de savoir si l'élément actuel est connecté à une entrée (ou est un entrée)."""
        return False

    # -------------
    # Methods
    # -------------
    def connector_bounding_rect(self):
        return self.connector_group.boundingRect()

    # -------------
    # Event Methods
    # -------------
    def connectorHoverEnterEvent(self, event: QGraphicsSceneHoverEvent) -> None:
        self._lbl.show()
        return super().hoverEnterEvent(event)

    def connectorHoverLeaveEvent(self, event: QGraphicsSceneHoverEvent) -> None:
        self._lbl.hide()
        return super().hoverLeaveEvent(event)

    def on_is_connected_changed(self):
        self._connector.background_color = Qt.GlobalColor.black if self.is_connected else Qt.GlobalColor.white

    def mousePressEvent (self, event: QGraphicsSceneMouseEvent) -> None :
        if (event.button() == Qt.LeftButton and self.isUnderMouse()) :
            start_drag(self, event)
            event.accept() 

        return super().mousePressEvent(event)

    def has_link(self) -> bool:
        return len(self._links) > 0
        
    def to_dict(self) -> dict:
        output_dict = {
            "text": self.text,
            "signal_indices": []
        }

        return output_dict

class OutputConnectionGroupModel(AbstractGraphicalElement):
    
    def __init__(self, name: str, is_infinite: bool, data_type: DataType, minimum_count: int, maximum_count: int, 
            default_count: int=0, group_position: QPointF = QPointF(), rotation: float = 0, text: str = '', parent: QGraphicsItem = None) -> None:
        super().__init__(name, group_position, rotation, text, parent)

        # Save group parameters
        self._name: str = name
        self._is_infinite = is_infinite
        self._data_type: DataType = data_type
        self._minimum_count: int = minimum_count if minimum_count >= 0 else 0
        self._maximum_count: int = maximum_count if maximum_count >= minimum_count else minimum_count+1

        # Init outputs list
        self._outputs: list[OutputConnectionModel] = []

        for _ in range(default_count):
            self.append_output()

    
    # -------------
    # Properties
    # -------------
    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, new_name: str) -> None:
        if new_name != self._name:
            self._name = new_name

    @property
    def is_infinite(self) -> bool:
        return self._is_infinite

    @is_infinite.setter
    def is_infinite(self, new_is_infinite: str) -> None:
        if new_is_infinite != self._is_infinite:
            self._is_infinite = new_is_infinite

    @property
    def data_type(self) -> DataType:
        return self._data_type

    @data_type.setter
    def data_type(self, new_data_type: str) -> None:
        if new_data_type != self._data_type:
            self._data_type = new_data_type

    @property
    def minimum_count(self) -> int:
        return self._minimum_count

    @minimum_count.setter
    def minimum_count(self, new_minimum_count: str) -> None:
        if new_minimum_count != self._minimum_count:
            self._minimum_count = new_minimum_count

    @property
    def maximum_count(self) -> int:
        return self._maximum_count

    @maximum_count.setter
    def maximum_count(self, new_maximum_count: str) -> None:
        if new_maximum_count != self._maximum_count:
            self._maximum_count = new_maximum_count

    @property
    def count(self) -> int:
        return len(self._outputs)

    @count.setter
    def count(self, new_count: str) -> None:
        if new_count != self._count:
            self._count = new_count

    @property
    def outputs(self) -> list[OutputConnectionModel]:
        return self._outputs

    @property
    def diagram_model (self) :
        return self.parent().diagram_model

        
    # -------------
    # Methods
    # -------------
    def connectors_bounding_rect(self):
        if len(self._outputs) == 0:
            return QRectF()
        elif len(self._outputs) == 1:
            return self._outputs[0].connector_bounding_rect()
        else:
            # Init connectors bounding rect
            connectors_bounding_rect = self._outputs[0].connector_bounding_rect()
            for output in self._outputs:
                connectors_bounding_rect = connectors_bounding_rect.united(output.connector_bounding_rect())

            return connectors_bounding_rect

    def append_output(self) -> OutputConnectionModel:
        # Add new graphical item in the group
        new_output = OutputConnectionModel(
            name = self._name, 
            data_type = self._data_type, 
            position = QPointF(0, len(self._outputs) * (AbstractConnectionModel.ItemHeight + 10)), 
            text = f"{self._name}_{len(self._outputs)}", 
            is_infinite = self._is_infinite, 
            parent=self
        )
        new_output.action_param_semaphore = False

        # Append output in the list
        self._outputs.append(new_output)

        return new_output

    def insert_output(self, output_index) -> OutputConnectionModel:
        # Recompute all outputs positions
        new_output = OutputConnectionModel(
            name = self._name, 
            data_type = self._data_type, 
            position = QPointF(0, len(self._outputs) * (AbstractConnectionModel.ItemHeight + 10)), 
            text = f"{self._name}_{len(self._outputs)}", 
            is_infinite = self._is_infinite, 
            parent=self
        )
        new_output.action_param_semaphore = False
        
        # Recompute all outputs positions
        self._outputs.insert(output_index, new_output)

        # Recompute all outputs positions
        self.recompute(from_index=output_index)

        return new_output

    @overload
    def remove_output(self, output_index: int) -> OutputConnectionModel:
        ...

    @overload
    def remove_output(self, output: OutputConnectionModel) -> OutputConnectionModel:
        ...

    def remove_output(self, output: Union[int, OutputConnectionModel]) -> OutputConnectionModel:
        if type(output) == int:
            # Get output index
            output_index = output
            
            # Get output item
            output = self._outputs[output_index]
        elif type(output) == OutputConnectionModel:
            # Get output index
            output_index = self._outputs.index(output)
            
            # Get output item
            output = self._outputs[output_index]
        else:
            LogManager().error(f"OutputConnectionGroupModel.remove_output(): given arg is not a supported type '{type(output)}' instead of 'int' or 'OutputConnectionModel'")
            return 
            
        # Remove item from list
        self._outputs.remove(output)

        # Remove from connection group
        output.setParentItem(None)

        # Delete graphical item
        output.deleteLater()

        # Recompute positions
        self.recompute(from_index=output_index)

    def clear_outputs(self):
        if self._is_infinite:
            # Remove only if this group is infinite
            while len(self._outputs) > 0:
                # Pop output from the list
                output = self._outputs.pop(0)

                # Remove graphical item
                output.deleteLater()

    def recompute(self, from_index: int = 0):
        # If index is out of list limit → don't do anything
        if from_index < 0 or from_index >= len(self._outputs):
            return

        # Recompute positions
        for output_index, output in enumerate(self._outputs[from_index:]):
            output.setPos(QPointF(0, output_index * (AbstractConnectionModel.ItemHeight + 10)))
            
    def to_dict(self) -> dict:
        outputs_group_dict = {
            "name": self.name,
            "isInfinite": self.is_infinite,
            "data_type": str(self.data_type),
            "outputs": [output.to_dict() for output in self._outputs]
        }

        return outputs_group_dict


from os import system
import time

import numpy as np
import matplotlib.pyplot as plt

from typing import List

from box_library.transfert_function_box import get_coeff_vector, euler
from muphyn.packages.unit_test.EDT_SFT_0048_2 import print_eq_laplace

# =========== Differential equation ===========

_ = system('cls')

wc = 20

num = '1'
denom = '1 1 1'

num_vect = get_coeff_vector(num)
denom_vect = get_coeff_vector(denom)

denom_order = len(denom_vect) - 1
num_order = len(num_vect) - 1
order = denom_order

# =========== Simulation ===========

fs = 100
dt = 1/fs
sim_time = 15
samples_number = int(fs * sim_time) + 2

print('=============================================================')
print('||                                                         ||')
print('||                      Simulation                         ||')
print('||                                                         ||')
print('|| - num order :', num_order, '                                        ||')
print('|| - denom order :', denom_order, '                                      ||')
print('|| - eq order :', order, '                                         ||')
print('||                                                         ||')
print('|| - simulation time :', sim_time, '                                 ||')
print('|| - sample frequency :', fs, '                              ||')
print('|| - dt :', dt, '                                           ||')
print('|| - samples number :', samples_number, '                               ||')
print('||                                                         ||')
print('|| - input : STEP                                          ||')
print('|| - eq : ', print_eq_laplace(num_vect, denom_vect))
print('||                                                         ||')
print('=============================================================')
print('')
print('')

num_vect.reverse()
denom_vect.reverse()

# time
t = [n * dt for n in range(samples_number)]

# input - STEP
u = [1 if t_ > 1 else 0 for t_ in t]

# output
last_u_derivatives : List[float] = [0 for num in num_vect]
last_y_derivatives : List[float] = [0 for denom in denom_vect]
last_y = 0
y = []


for i, t_ in enumerate(t) :
#for i in range(12) :

    current_y = 0
    current_u_derivatives = []
    current_y_derivatives = []

    # u's
    for j, coeff_u in enumerate(num_vect) :
        
        if j == 0 :
            current_y += coeff_u * u[i]
            current_u_derivatives.append(u[i])

        else:
            derivative = (current_u_derivatives[j - 1] - last_u_derivatives[j - 1]) / dt

            current_y += coeff_u * derivative
            current_u_derivatives.append(derivative)

    # y's 
    if order > 0 :
        
        for j in range(order) : 

            coeff_y = denom_vect[j]
            
            if j == 0 :
                current_y -= coeff_y * last_y
                current_y_derivatives.append(last_y)

            else :
                derivative = (current_y_derivatives[j - 1] - last_y_derivatives[j - 1]) / dt

                current_y -= coeff_y * derivative
                current_y_derivatives.append(derivative)
        
        current_y /= denom_vect[order]

        for j in range(order).__reversed__() :
            print('time :', t_, ' - order :', j, ' - derivative :', current_y_derivatives[j])
            current_y *= dt
            current_y += current_y_derivatives[j]
            

    else :
        current_y = current_y / denom_vect[order]

    """
    if i >= 2 :
        #current_y -= denom_vect[0] * y[i - 1]
        #current_y -= denom_vect[1] * (current_y_derivatives[0] - last_y_derivatives[0]) / dt
        #current_y /= denom_vect[2]
        current_y *= dt
        current_y += (y[i - 1] - y[i - 2]) / dt
        current_y *= dt 
        current_y += y[i - 1]
        
    elif i == 1 :
        #current_y -= denom_vect[0] * y[i - 1]
        #current_y -= denom_vect[1] * (current_y_derivatives[0]) / dt
        #current_y /= denom_vect[2]
        current_y *= dt
        current_y += ((y[i - 1]) / dt)
        current_y *= dt 
        current_y += y[i - 1]

    else :
        #current_y /= denom_vect[2]
        current_y *= dt
        current_y *= dt
    """

    last_u_derivatives = current_u_derivatives
    last_y_derivatives = current_y_derivatives
    last_y = current_y
    
    y.append(current_y)

print('at', (len(y) - 1) / fs, 's :', y[len(y) - 1])

# =========== Personnal notes ===========

print('')
print(' =====> Résultats <=====')
print('')
print('date : 22/02/2022')
print('')
print('Les modifications apportées et trouvées permettent ')

# =========== Plot results ===========

t_renderable = t[0 : len(y)]
u_renderable = u[0 : len(y)]

plt.figure()
num_vect.reverse()
denom_vect.reverse()
plt.title(print_eq_laplace(num_vect, denom_vect))
#plt.plot(t_renderable, u_renderable, label="u(t)")
plt.plot(t_renderable, y, label="y(t)")
plt.legend()
plt.grid()
plt.xlabel('Time')
plt.show()


#-----------------------------------
# Imports
#-----------------------------------

from typing import Any

from muphyn.packages.interface.graphical_actions.abstract_graphical_action import AbstractGraphicalAction

#-----------------------------------
# Class
#-----------------------------------

class AbstractDiagramAction (AbstractGraphicalAction) :
    """Est la classe abstraite commune aux actions modifiant les informations contenus dans les modèles de diagram."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, diagram_model : Any) :

        AbstractGraphicalAction.__init__(self)
        self._diagram_model = diagram_model

    # -------------
    # Properties
    # -------------

    @property
    def diagram_model (self) -> Any :
        """Permet de récuperer le modèle diagram qu'il faut modifier avec les actions."""
        return self._diagram_model


#-----------------------------------
# Imports
#-----------------------------------

from typing import Iterable

from PyQt5.QtGui import QColor
from PyQt5.QtCore import QPointF, QSizeF, Qt
from PyQt5.QtWidgets import QGraphicsItem

from muphyn.packages.core.plci_core_data import Data
from muphyn.packages.core.plci_core_data_type import DataType
from muphyn.packages.interface.models.graphical_models.abstract_box_IO_model import AbstractBoxIOModel
from muphyn.packages.interface.models.signals_model.input_connection_model import InputConnectionModel
from muphyn.packages.interface.models.signals_model.signal_link_model import SignalLinkModel

#-----------------------------------
# Class
#-----------------------------------

class BoxOutputModel (AbstractBoxIOModel) :
    """Est la classe des sorties des boxes composite."""
    
    # -------------
    # Constructors
    # -------------

    def __init__ (self, name : str, data_type : DataType, position : QPointF, size : QSizeF,
                 rotation : float = 0.0, links : Iterable[SignalLinkModel] = [], text : str = '',
                 color: QColor = Qt.GlobalColor.black, parent : QGraphicsItem = None) :

        AbstractBoxIOModel.__init__(self, name, data_type, position, size, rotation, text, parent)

        self._inputs.append(InputConnectionModel('', data_type, QPointF(0, 0), '', color, self))
    
    # -------------
    # Properties
    # -------------

    @property
    def default_value (self) -> Data :
        """Permet de récuperer la valeur par défaut du lien."""
        return self._default_value

    @default_value.setter
    def default_value (self, default_value_ : Data) :
        """Permet de modifier la valeur par défaut du lien."""

        if default_value_ is None :
            self._default_value = Data(self.data_type)

        elif not default_value_._data_type == self.data_type :
            self._default_value = Data(self.data_type)
        
        else :
            self._default_value = default_value_


# General imports
import os
import re
from dataclasses import dataclass

# PyQt imports
from PyQt5.QtCore import QPoint, QRect, QSize
from PyQt5.QtGui import QColor

@dataclass
class OpenModelicaInput:
    type: str
    name: str
    value: float = 0

    def __str__(self):
        return f"{self.type} {self.name} (start = {self.value})"

@dataclass
class OpenModelicaOutput:
    type: str
    name: str

    def __str__(self):
        return f"{self.type} {self.name}"

@dataclass
class OpenModelicaTransformation:
    origin: QPoint
    extent: list[QPoint]
    rotation: int

    def boundingRect(self):
        return QRect(self.topLeft(), self.size())

    def size(self) -> QSize:
        if self.extent is not None:
            return QSize(self.extent[1].x() - self.extent[0].x(), self.extent[1].y() - self.extent[0].y())
        else:
            return QSize()

    def topLeft(self) -> QPoint:
        topLeft = QPoint()

        if self.origin is not None:
            topLeft += self.origin

        if self.extent is not None:
            topLeft += self.extent[0]

        return topLeft

    def transformOrigin(self) -> QPoint:
        return self.boundingRect().center()

    def __str__(self) -> str:
        string_list = []
        
        if self.origin is not None:
            string_list.append(f"origin = {{{self.origin.x()}, {self.origin.y()}}}")
            
        if self.extent is not None:
            string_list.append(f"extent = {{{{{self.extent[0].x()}, {self.extent[0].y()}}}, {{{self.extent[1].x()}, {self.extent[1].y()}}}}}")
            
        if self.rotation is not None:
            string_list.append(f"rotation = {self.rotation}")
            
        return ", ".join(string_list)
    
@dataclass
class OpenModelicaComponentAnnotation:
    visible: bool
    transformation: OpenModelicaTransformation = None

    def __str__(self) -> str:
        return f"Placement(visible = {str(self.visible).lower()}, transformation({self.transformation}))"
    
@dataclass
class OpenModelicaComponent:
    library: str
    name: str
    annotation: OpenModelicaComponentAnnotation = None

    def boundingRect(self) -> QRect:
        if self.annotation is not None and self.annotation.transformation is not None:
            return self.annotation.transformation.boundingRect()
        else:
            return QRect()

    def rotation(self) -> int:
        if self.annotation is not None and self.annotation.transformation is not None:
            return self.annotation.transformation.rotation
        else:
            return 0

    def topLeft(self) -> QPoint:
        if self.annotation is not None and self.annotation.transformation is not None:
            return self.annotation.transformation.topLeft()
        else:
            return QPoint()

    def transformOrigin(self) -> QPoint:
        if self.annotation is not None and self.annotation.transformation is not None:
            return self.annotation.transformation.transformOrigin()
        else:
            return QPoint()

    def size(self) -> QSize:
        if self.annotation is not None and self.annotation.transformation is not None:
            return self.annotation.transformation.size()
        else:
            return QSize()

    def __str__(self):
        return f"{self.library} {self.name} annotation({self.annotation})"

@dataclass
class OpenModelicaLine:
    points: list[QPoint]
    color: QColor

    def __str__(self) -> str:
        points_str = ", ".join([f"{{{point.x()}, {point.y()}}}" for point in self.points])
        color_str = f"{{{self.color.red()}, {self.color.green()}, {self.color.blue()}}}"
        return f"points = {{{points_str}}}, color = {color_str}"

@dataclass
class OpenModelicaConnectionAnnotation:
    line: OpenModelicaLine

    def __str__(self) -> str:
        return f"Line({self.line})"

@dataclass
class OpenModelicaConnection:
    node1_name: str
    node2_name: str
    annotation: OpenModelicaConnectionAnnotation = None

    def points(self) -> list[QPoint]:
        if self.annotation is not None and self.annotation.line is not None:
            return self.annotation.line.points

    def __str__(self) -> str:
        return f"connect({self.node1_name}, {self.node2_name}) annotation({self.annotation})"


class OpenModelicaModelParser:

    # General Open Modelica model regex
    GlobalModelPattern = r"\Amodel\s(?P<model_name_1>\w+){1}\s(?P<model_content>.+);end\s(?P<model_name_2>\w+){1};\Z"

    # Declaration Part - Variables & Component patterns
    InputPattern = r"(?i)input\s*(?P<type>\w+)\s*(?P<name>\w+)\s*(?:\(start\s*\=\s*(?P<start_value>[0-9]+)\))?"
    OutputPattern = r"(?i)output\s*(?P<type>\w+)\s*(?P<name>\w+)"
    ComponentPattern = r"(?i)(?P<library>\w+(?:\.\w+)+)\s+(?P<name>\w+)\s*(?:\((?P<parameters>.*)\))?\s+annotation\s*\(\s*(?P<annotation>.*)\)"
    ComponentParameterPattern = r"(?i)(?P<name>\w+)\s*(?:\(\s*(?P<added_parameter>\w+)\s*=\s*(?P<added_parameter_value>true|false|[+-]?[0-9]+(?:\.[0-9]*)?|\"[a-z]*\")\))?\s*(?:=\s*(?P<value>true|false|[+-]?[0-9]+(?:\.[0-9]*)?|\"[a-z]*\"))?"
    ComponentAnnotationPattern = r"(?i)Placement\s*\(\s*(?:visible\s*=\s*(?P<visible>true|false),?)?\s*,?\s*\s*(?P<transformation>transformation\s*\(.*\))?\s*\)\s*"

    # Equation part - Equations, connections & annotation
    ConnectionPattern = r"(?i)connect\(\s*(?P<node_1>.*)\s*,\s*(?P<node_2>.*)\s*\)\s+annotation\(\s*(?P<annotation>.*)\s*\)"
    ConnectionAnnotationPattern = r"(?i)line\s*\(\s*points\s*=\s*{\s*(?P<points>(?:\s*{\s*[+-]?[0-9]+\s*,\s*[+-]?[0-9]+\s*}\s*,?)+)\s*}\s*,\s*color\s*=\s*(?P<color>{\s*(?P<red>[0-9]+)\s*,\s*(?P<green>[0-9]+)\s*,\s*(?P<blue>[0-9]+)})\s*\)"
    
    # Data extractor regexes
    ColorRegex = r"{\s*(?P<red>[0-9]{1,3})\s*,\s*(?P<green>[0-9]{1,3})\s*,\s*(?P<blue>[0-9]{1,3})\s*}"
    PointRegex = r"{\s*(?P<x>[+-]?[0-9]+)\s*,\s*(?P<y>[+-]?[0-9]+)\s*}"
    TransformationPattern = r"(?i)transformation\s*\(\s*(origin\s*=\s*\{\s*(?P<origin_x>[+-]?[0-9]*)\s*,\s*(?P<origin_y>[+-]?[0-9]*)\s*\}\s*,?)?\s*(extent\s*=\s*\{\{(?P<extent_0_x>[+-]?[0-9]*)\s*,\s*(?P<extent_0_y>[+-]?[0-9]*)\s*\}\s*,\s*\{(?P<extent_1_x>[+-]?[0-9]*)\s*,\s*(?P<extent_1_y>[+-]?[0-9]*)\s*\}\}\s*,?)?\s*(rotation\s*=\s*(?P<rotation>[+-]?[0-9]*),?)?\s*\)\s*"

    def __init__(self, om_file_path: str) -> None:
        # Test if file exists
        if not os.path.exists(om_file_path):
            raise(FileNotFoundError(f"File not found: {om_file_path}"))

        # Path
        self.path = om_file_path

        # Read file
        file_content = ""
        with open(self.path, mode='r') as f:
            file_content = f.read().strip()

        # Test if empty file
        if file_content == "":
            raise(Exception("Empty file"))

        # Parse file
        self.inputs: list[OpenModelicaInput] = []
        self.outputs: list[OpenModelicaOutput] = []
        self.components: list[OpenModelicaComponent] = []
        self.connections: list[OpenModelicaConnection] = []
        self.other_init_rows = []
        self.other_equation_rows = []
        self.parse_file(file_content)

    def parse_file(self, file_content):
        content = file_content.replace('\n', '')

        if re.match(OpenModelicaModelParser.GlobalModelPattern, content):
            for match in re.finditer(OpenModelicaModelParser.GlobalModelPattern, content):
                # Get component name
                model_name_1 = match.groupdict()["model_name_1"]

                # Get component name
                model_name_2 = match.groupdict()["model_name_2"]

                if model_name_1 != model_name_2:
                    raise(Exception("Model file has wrong model construction: starting model name ({model_name_1}) is not the same as the end ({model_name_2})"))

                self.model_name = model_name_1

                # Get content
                model_content = match.groupdict()["model_content"].strip()

                # Separate content into parts -> (items declaration, equations...)
                init_content, equations_content = model_content.split("equation", maxsplit=2)

                # Declaration handling
                init_rows = [row.strip() for row in init_content.split(";")]
                
                for init_row in init_rows:
                    # Input
                    if init_row == "":
                        pass
                    elif re.match(OpenModelicaModelParser.InputPattern, init_row):
                        for input_match in re.finditer(OpenModelicaModelParser.InputPattern, init_row):
                            input_match_dict = input_match.groupdict()
                            input_type = input_match_dict["type"]
                            input_name = input_match_dict["name"]
                            input_value = 0 if not "start_value" in input_match_dict else input_match_dict["start_value"]

                            self.inputs.append(OpenModelicaInput(input_type, input_name, input_value))

                    # Output
                    elif re.match(OpenModelicaModelParser.OutputPattern, init_row):
                        for output_match in re.finditer(OpenModelicaModelParser.OutputPattern, init_row):
                            output_type = output_match.groupdict()["type"]
                            output_name = output_match.groupdict()["name"]

                            self.outputs.append(OpenModelicaOutput(output_type, output_name))

                    # Component
                    elif re.match(OpenModelicaModelParser.ComponentPattern, init_row):
                        for component_match in re.finditer(OpenModelicaModelParser.ComponentPattern, init_row):
                            component_library = component_match.groupdict()["library"]
                            component_name = component_match.groupdict()["name"]
                            component_annotation = component_match.groupdict()["annotation"]

                            component = OpenModelicaComponent(component_library, component_name)

                            for annotation_match in re.finditer(OpenModelicaModelParser.ComponentAnnotationPattern, component_annotation):
                                annotation_visible = bool(annotation_match.groupdict()["visible"])
                                annotation_transformation = annotation_match.groupdict()["transformation"]

                                annotation = OpenModelicaComponentAnnotation(annotation_visible)
                                if annotation_transformation is not None: 
                                    for transformation_match in re.finditer(OpenModelicaModelParser.TransformationPattern, annotation_transformation):
                                        transformation_origin_x = int(transformation_match.groupdict()["origin_x"])
                                        transformation_origin_y = int(transformation_match.groupdict()["origin_y"])
                                        transformation_extent_0_x = int(transformation_match.groupdict()["extent_0_x"])
                                        transformation_extent_0_y = int(transformation_match.groupdict()["extent_0_y"])
                                        transformation_extent_1_x = int(transformation_match.groupdict()["extent_1_x"])
                                        transformation_extent_1_y = int(transformation_match.groupdict()["extent_1_y"])
                                        transformation_rotation = int(transformation_match.groupdict()["rotation"])

                                        origin_coords = [transformation_origin_x, transformation_origin_y]
                                        extent_coords = [transformation_extent_0_x, transformation_extent_0_y, transformation_extent_1_x, transformation_extent_1_y]

                                        # Add transformation to annotation
                                        annotation.transformation = OpenModelicaTransformation(
                                            origin=None if any(coord is None for coord in origin_coords) else QPoint(transformation_origin_x, transformation_origin_y),
                                            extent=None if any(coord is None for coord in extent_coords) else [QPoint(transformation_extent_0_x, transformation_extent_0_y), QPoint(transformation_extent_1_x, transformation_extent_1_y)],
                                            rotation = transformation_rotation
                                        )

                                # Add annotation to component
                                component.annotation = annotation

                            self.components.append(component)
                    else:
                        self.other_init_rows.append(init_row)

                # Equations handling
                equation_rows = [row.strip() for row in equations_content.split(";")]

                connections = []
                for equation_row in equation_rows:
                    if equation_row == "":
                        pass
                    elif re.match(OpenModelicaModelParser.ConnectionPattern, equation_row):
                        for connection_match in re.finditer(OpenModelicaModelParser.ConnectionPattern, equation_row):
                            connection_node_1 = connection_match.groupdict()["node_1"]
                            connection_node_2 = connection_match.groupdict()["node_2"]
                            connection_annotation = connection_match.groupdict()["annotation"]

                            connection = OpenModelicaConnection(connection_node_1, connection_node_2)
                            for connection_annotation_match in re.finditer(OpenModelicaModelParser.ConnectionAnnotationPattern, connection_annotation):

                                # Extract points
                                points_string = connection_annotation_match.groupdict()["points"]
                                points = [QPoint(int(point_match.groupdict()["x"]), int(point_match.groupdict()["y"])) for point_match in re.finditer(OpenModelicaModelParser.PointRegex, points_string)]

                                # Extract color
                                red = int(connection_annotation_match.groupdict()["red"])
                                green = int(connection_annotation_match.groupdict()["green"])
                                blue = int(connection_annotation_match.groupdict()["blue"])

                                line = OpenModelicaLine(points, QColor(red, green, blue))

                                connection.annotation = OpenModelicaConnectionAnnotation(line)

                            connections.append(connection)
                    else:
                        self.other_equation_rows.append(equation_row)

                # Save to object
                self.connections = connections

    def save_file_as(self, output_file_path: str):
        if output_file_path == "":
            raise(ValueError("Empty string given for output file path"))
        else:
            dirname = os.path.dirname(output_file_path)
            if not os.path.exists(dirname):
                os.makedirs(dirname)

            # Build output model
            file_content = f"model {self.model_name}\n"

            # Inputs
            file_content += '\n'.join([f"  input {input_});" for input_ in self.inputs])
            file_content += '\n'

            # Outputs
            file_content += '\n'.join([f"  output {output};" for output in self.outputs])
            file_content += '\n'

            # Components
            file_content += '\n'.join([f"  {component});" for component in self.components])
            file_content += '\n'

            # Other init rows
            file_content += '\n'.join([f"{other_init_row};" for other_init_row in self.other_init_rows])
            file_content += '\n'

            # Equation part
            file_content += "equation\n"
            
            # Connection
            file_content += '\n'.join([f"  {connection};" for connection in self.connections])

            # Other init rows
            file_content += '\n'.join([f"  {other_equation_row};" for other_equation_row in self.other_equation_rows])
            file_content += '\n'

            # End of file
            file_content += f"end {self.model_name};"

            # with open(output_file_path, "w") as f:
            #     f.write(file_content)

    def set_input_value(self, input_name: str, input_value: float):
        for input_ in self.inputs:
            if input_.name == input_name:
                input_.name = input_value

    def __str__(self):
        to_print = ["* Inputs *"]
        for input_ in self.inputs:
            to_print.append(f"\t - {input_}")

        to_print.append("* Outputs *")
        for output in self.outputs:
            to_print.append(f"\t - {output}")

        to_print.append("* Components *")
        for component in self.components:
            to_print.append(f"\t - {component}")

        to_print.append("* Connections *")
        for connection in self.connections:
            to_print.append(f"\t - {connection}")

        return "\n".join(to_print)
        


from enum import Enum
from PyQt5.QtCore import QLineF, QPointF, QSizeF
from PyQt5.QtWidgets import QGraphicsItem
from PyQt5.QtGui import QColor

from basic_shape.shapes import Square, Line
from basic_shape.grouped_shapes import GroupedShapes



class ElectricalConnector(GroupedShapes):

    # Elecrtical node type
    class ElectricalConnectorType(Enum):
        Positive = 1
        Neutral = 0
        Negative = -1

    # Node Size
    node_side_size = 10
    connector_line_length = 10
    connector_size = QSizeF(node_side_size + connector_line_length, node_side_size)

    # Node Color    
    border_node_color = QColor(0, 0, 255)
    positive_node_color = border_node_color
    negative_node_color = QColor(255, 255, 255)

    def __init__(self, type: ElectricalConnectorType, position: QPointF = QPointF(0, 0), parent: QGraphicsItem = None) -> None:

        super().__init__(position, parent=parent)

        # Select Color
        border_color = ElectricalConnector.border_node_color
        background_color = ElectricalConnector.get_color_by_type(type)

        # Shapes
        offset = position + QPointF(0, -ElectricalConnector.node_side_size/2)

        # Square
        square = Square(
            QPointF(0, 0) + offset, 
            ElectricalConnector.node_side_size, 
            border_color=border_color, border_width=2, fill_color=background_color,
            parent=self
        )

        # Line
        start_point = QPointF(ElectricalConnector.node_side_size, ElectricalConnector.node_side_size/2) + offset
        end_point = QPointF(ElectricalConnector.node_side_size + ElectricalConnector.connector_line_length, ElectricalConnector.node_side_size/2) + offset
        line = Line(
            QLineF(start_point, end_point),
            line_color=border_color, line_width=2,
            parent=self
        )

        # self.add_shapes([square, line])


    # --------------
    # Static Methods
    # --------------
    @staticmethod
    def get_color_by_type(type: ElectricalConnectorType) -> QColor:
        color_switcher = {
            ElectricalConnector.ElectricalConnectorType.Positive: ElectricalConnector.positive_node_color,
            ElectricalConnector.ElectricalConnectorType.Neutral: ElectricalConnector.positive_node_color,
            ElectricalConnector.ElectricalConnectorType.Negative: ElectricalConnector.negative_node_color
        }

        return color_switcher[type]


#-----------------------------------
# Imports
#-----------------------------------

# Project Imports
from muphyn.packages.interface.graphical_actions.abstract_diagram_action import AbstractDiagramAction
from muphyn.packages.interface.models.graphical_models.abstract_graphical_element import AbstractGraphicalElement

#-----------------------------------
# Class
#-----------------------------------

class AbstractUniqueElementDiagramAction (AbstractDiagramAction) :
    """Est la classe abstraite commune aux actions modifiant les informations contenus dans un modèle unique d'un diagrame."""
    
    # -------------
    # Constructors
    # -------------

    def __init__ (self, graphical_element : AbstractGraphicalElement) :

        AbstractDiagramAction.__init__(self, graphical_element.diagram_model)
        self._graphical_index = graphical_element.graphical_index
        
    # -------------
    # Properties
    # -------------
    
    @property
    def graphical_element (self) :
        """Permet de récuperer l'élément graphique qui doit être modifiée."""
        return self._diagram_model.get_element_by_graphical_index(self._graphical_index)

    @property
    def graphical_index (self) :
        """Permet de récuperer l'index de l'élément graphique qui doit être modifié."""
        return self._graphical_index


#-----------------------------------
# Imports
#-----------------------------------

from typing import List, Dict

from muphyn.packages.interface.graphical_actions.abstract_unique_element_diagram_action import AbstractUniqueElementDiagramAction
from muphyn.packages.interface.models.graphical_models.abstract_box_model import AbstractBoxModel
from muphyn.packages.interface.models.signals_model.output_connection_model import OutputConnectionModel
from muphyn.packages.interface.graphical_actions import parser_decode as parser_decode
from muphyn.packages.interface.graphical_actions import parser_encode as parser_encode

#-----------------------------------
# Class
#-----------------------------------

class DiagramRemoveOutputsFromBoxAction (AbstractUniqueElementDiagramAction) :
    """Est l'action permettant de supprimer les sorties d'une box."""

    # -------------
    # Constructors
    # -------------
    
    def __init__ (self, box_element : AbstractBoxModel, output_group_name: str, outputs : List[OutputConnectionModel]) :

        AbstractUniqueElementDiagramAction.__init__(self, box_element)

        self._reconstructors : List[Dict] = []
        self._output_group_name: str = output_group_name
        self._outputs: list[OutputConnectionModel] = outputs
        
    # -------------
    # Methods
    # -------------
    def do (self):

        box_model: AbstractBoxModel = self.graphical_element

        if box_model is None : 
            return

        for output in self._outputs:
            output_dict : Dict = {"graphical_index": output.graphical_index}
            current_output : Dict = parser_encode.box_output(output, output_dict)
            current_output['signals'] = []

            for signal in output.links :
                current_output['signals'].append(parser_encode.link(signal))
                signal.unbind()
                self.diagram_model.remove_element(signal)

            output._links.clear()
            
            box_model.remove_output(self._output_group_name, output)

            self._reconstructors.append(output_dict)

    def undo (self) :
        
        box_model : AbstractBoxModel = self.graphical_element 

        if box_model is None : 
            return

        self._reconstructors.reverse()
        for output_dict in self._reconstructors :
            parser_decode.box_output(output_dict, box_model)

            for signal_dict in output_dict['signals'] : 
                parser_decode.link(signal_dict, self.diagram_model)


#-----------------------------------
# Imports
#-----------------------------------

from PyQt5.QtCore import Qt, QCoreApplication
from PyQt5.QtWidgets import QListWidgetItem, QPushButton, QWidget, QLabel, QHBoxLayout, QVBoxLayout, QFormLayout, QListWidget, QAbstractItemView
from muphyn.packages.core.plci_core_data_type import DataType
from muphyn.packages.core.utils.log_manager import LogManager
from muphyn.packages.interface.models.graphical_models.box_model import BoxModel
from muphyn.packages.interface.properties_pages.abstract_properties_editor import AbstractPropertiesEditor
from muphyn.packages.interface.properties_pages.infinite_io_properties_item import InifiniteIOPropertiesItem

#-----------------------------------
# Class
#-----------------------------------

class AbstractInfiniteIOPropertiesEditor (AbstractPropertiesEditor) :
    """Est la classe abstraite commune aux panneaux de propriétés qui permets d'afficher et modifier les entrées/sorties infinies."""

    # -------------
    # Constants
    # -------------
    ItemHeight: int = InifiniteIOPropertiesItem.ItemHeight * 3
    
    # -------------
    # Constructors
    # -------------

    def __init__ (self, box_model : BoxModel, group_name : str, data_type : DataType, minimum_count: int, maximum_count: int) :
        
        AbstractPropertiesEditor.__init__(self, box_model)

        # Box Model
        self._box_model = box_model

        # Get input group name
        self._group_name = group_name

        # Data Type
        self._data_type = data_type

        # Count
        self._minimum_count = minimum_count
        self._maximum_count = maximum_count

        self._lbl_connection_name_value.setText(group_name)
        self._lbl_data_type_value.setText(self._data_type.__str__())

    # -------------
    # Properties
    # -------------

    @property
    def box_model (self) -> BoxModel :
        """Permet de récuperer le modèle de box modifié par le pannel actuel."""
        return self._box_model
    
    @property
    def group_name (self) -> str :
        """Permet de récuperer le nom de la connexion."""
        return self._group_name

    @property 
    def data_type (self) -> DataType : 
        """Permet de récuperer le type de connexion."""
        return self._data_type

    @property 
    def minimum_count (self) -> int :
        return self._minimum_count

    @property 
    def maximum_count (self) -> int :
        return self._maximum_count

    # -------------
    # Methods
    # -------------

    def init_ui (self) -> None :
        
        self._pnl_form : QWidget = QWidget()
        self._lyt_form : QFormLayout = QFormLayout()
        self._lbl_connection_name : QLabel = QLabel()
        self._lbl_connection_name_value : QLabel = QLabel()
        self._lyt_form.addRow(self._lbl_connection_name, self._lbl_connection_name_value)

        self._lbl_data_type : QLabel = QLabel()
        self._lbl_data_type_value : QLabel = QLabel()
        self._lyt_form.addRow(self._lbl_data_type, self._lbl_data_type_value)

        self._lbl_connection_mode : QLabel = QLabel()
        self._lbl_connection_mode_value : QLabel = QLabel()
        self._lyt_form.addRow(self._lbl_connection_mode, self._lbl_connection_mode_value)

        self._pnl_button_holder : QWidget = QWidget()

        self._btn_add : QPushButton = QPushButton()
        self._btn_add.setBaseSize(100, 22)
        self._btn_add.clicked.connect(self.button_add_click)

        self._btn_remove : QPushButton = QPushButton()
        self._btn_remove.setBaseSize(100, 22)
        self._btn_remove.clicked.connect(self.button_remove_click)

        self._lyt_button_holder : QHBoxLayout = QHBoxLayout()
        self._lyt_button_holder.addWidget(self._btn_add)
        self._lyt_button_holder.addWidget(self._btn_remove)
        self._lyt_button_holder.setAlignment(Qt.AlignRight)

        self._pnl_button_holder.setLayout(self._lyt_button_holder)
        self._lyt_form.addRow(QLabel(), self._pnl_button_holder)

        self._pnl_form.setLayout(self._lyt_form)
        self.layout().addWidget(self._pnl_form)

        self._tbl_connection : QListWidget = QListWidget()
        self._tbl_connection.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.layout().addWidget(self._tbl_connection)
        self.layout().setContentsMargins(0, 0, 0, 0)


    def translate_ui (self) -> None :

        self._lbl_connection_name.setText(QCoreApplication.translate(self.objectName(), u'Name : ', None))
        self._lbl_data_type.setText(QCoreApplication.translate(self.objectName(), u'Data type : ', None))
        self._lbl_connection_mode.setText(QCoreApplication.translate(self.objectName(), u'Connexion mode : ', None))
        self._btn_add.setText(QCoreApplication.translate(self.objectName(), u"+", None))
        self._btn_remove.setText(QCoreApplication.translate(self.objectName(), u"-", None))

    def button_add_click (self) -> None : 
        """Est la méthode appelée lorsque l'utilisateur veut cliquer sur le bouton ajouter."""
        LogManager().debug('btn add clicked')
        
    def button_remove_click (self) -> None : 
        """Est la méthode appelée lorsque l'utilisateur veut cliquer sur le bouton ajouter."""
        LogManager().debug('btn remove clicked')

    def create_layout (self) -> None : 
        """Permet de créer le layout pour l'affichage actuel."""
        return QVBoxLayout(self)

        
    def recompute_connection_numbers (self) -> None : 
        
        for input_index in range(self._tbl_connection.count()) :
            item = self._tbl_connection.item(input_index)
            item_widget : InifiniteIOPropertiesItem = self._tbl_connection.itemWidget(item)
            item_widget.number = input_index + 1



#-----------------------------------
# Imports
#-----------------------------------

from typing import List

#-----------------------------------
# Functions
#-----------------------------------

def _init_constant_box (box, simulation_params) -> None :
    
    if not 'Constant Value' in box :
        box['Constant Value'] = 1

def _function_constant_box (box, event_) -> List :
    
    v = box['Constant Value']
    events : List = []
    
    for output in box.outputs :
        events.append(box.construct_signal_event(output, v))

    return events


#-----------------------------------
# Imports
#-----------------------------------

import matplotlib.pylab as plt
from typing import List

from muphyn.packages.core.plci_core_box import Box
from muphyn.packages.core.plci_core_scheduler_event import SchedulerEvent
from muphyn.packages.core.plci_core_scheduler_params import SchedulerParams
from muphyn.packages.interface.models.graphical_models.box_model import BoxModel

from muphyn.packages.interface.dialogs.graph_dialog import GraphWindow

#-----------------------------------
# Functions
#-----------------------------------

def _init_graph_box (box: BoxModel, simulation_params: SchedulerParams) -> None :

    box['data_x'] = []
    box['data_y'] = {input_.input_name: [] for input_ in box.inputs}

    box['point_count'] = 0

    if not 'title' in box :
        box['title'] = str(box.name) + " " + str(box.index)
    
    if not 'label_x' in box :
        box['label_x'] = ' '

    if not 'label_y' in box :
        box['label_y'] = ' '

    if not 'start_time' in box :
        box['start_time'] = 1.0

    if not 'stop_time' in box : 
        box['stop_time'] = 100.0
    
    box['window'] = None

def _function_graph_box (box: Box, event_: SchedulerEvent) -> List:
    
    if box['copy_timing_from_simulation'] is True or (event_.timing > box['start_time'] and event_.timing < box['stop_time']) :
        box['data_x'].append(event_.timing)
        
        for input_ in box.inputs :
            box['data_y'][input_.input_name].append(input_.value)

        box['point_count'] = box['point_count'] + 1

    return []

def _end_graph_box (box: Box) -> None :
    box['window'] = GraphWindow(box = box)
    box['window'].show()


import sys
import matplotlib as mpl
from matplotlib.backends.backend_agg import FigureCanvasAgg
from PyQt5 import QtGui, QtCore

def mathTex_to_QPixmap(mathTex, fs = 12):

    #---- set up a mpl figure instance ----

    fig = mpl.figure.Figure()
    fig.patch.set_facecolor('none')
    fig.set_canvas(FigureCanvasAgg(fig))
    renderer = fig.canvas.get_renderer()

    #---- plot the mathTex expression ----

    ax = fig.add_axes([0, 0, 1, 1])
    ax.axis('off')
    ax.patch.set_facecolor('none')
    t = ax.text(0, 0, mathTex, ha='left', va='bottom', fontsize=fs)

    #---- fit figure size to text artist ----

    fwidth, fheight = fig.get_size_inches()
    fig_bbox = fig.get_window_extent(renderer)

    text_bbox = t.get_window_extent(renderer)

    tight_fwidth = text_bbox.width * fwidth / fig_bbox.width
    tight_fheight = text_bbox.height * fheight / fig_bbox.height

    fig.set_size_inches(tight_fwidth, tight_fheight)

    #---- convert mpl figure to QPixmap ----

    buf, size = fig.canvas.print_to_buffer()
    qimage = QtGui.QImage.rgbSwapped(QtGui.QImage(buf, size[0], size[1],
                                                  QtGui.QImage.Format_ARGB32))
    qpixmap = QtGui.QPixmap(qimage)

    return qpixmap



#-----------------------------------
# Imports
#-----------------------------------

from PyQt5 import QtGui
from PyQt5.QtCore import QCoreApplication, QRect, Qt
from PyQt5.QtWidgets import QLabel, QVBoxLayout

from muphyn.packages.interface.properties_pages.abstract_properties_editor import AbstractPropertiesEditor

#-----------------------------------
# Class
#-----------------------------------

class UnknownPropertiesEditor (AbstractPropertiesEditor) :
    """Est la page affichées quand aucune autre page de propriétés ne peut être affichées."""
        
    # -------------
    # Constructors
    # -------------
    
    def __init__ (self) :

        AbstractPropertiesEditor.__init__(self, None)

    # -------------
    # Methods
    # -------------

    def resizeEvent (self, event : QtGui.QResizeEvent) -> None :
        self.redraw_ui()
        return super().resizeEvent(event)

    def init_ui (self) :

        if not self.objectName():
            self.setObjectName(u"pnl_unknown_properties")

        self._lbl_unknown : QLabel = QLabel(self)
        self._lbl_unknown.setAlignment(Qt.AlignCenter)
    
        self.redraw_ui()

    def translate_ui (self) -> None :

        self._lbl_unknown.setText(QCoreApplication.translate(self.objectName(), u"No selected item", None))

    def redraw_ui (self) -> None :

        self._lbl_unknown.setGeometry(QRect(0, 10, self.width(), 35))

        self._height = 40

    def create_layout (self) -> None : 
        """Permet de créer le layout pour l'affichage actuel."""
        return QVBoxLayout(self)


#-----------------------------------
# Imports
#-----------------------------------

# General Imports
import yaml
from datetime import date
from typing import Dict

# Project Imports
from muphyn.packages.interface.files.abstract_exporter import AbstractExporter
from muphyn.packages.interface.models.graphical_models.box_model import BoxModel
from muphyn.packages.interface.models.editable_models.simulation_model import SimulationModel
from muphyn.packages.interface.models.signals_model.abstract_signal_model import AbstractSignalModel
from muphyn.packages.interface.models.links_model.abstract_link_model import AbstractLinkModel
from muphyn.packages.interface.models.signals_model.input_connection_model import InputConnectionModel
from muphyn.packages.interface.models.signals_model.output_connection_model import OutputConnectionModel

def export_box (box : BoxModel) -> Dict :
    """Est la méthode appelée pour créer un dictionnaire contenant les données d'une box."""

    box_dict = {}

    box_dict['name'] = box.name
    box_dict['library'] = box.library
    box_dict['text'] = box.text

    box_dict['geometry'] = {}
    box_dict['geometry']['x'] = box.x()
    box_dict['geometry']['y'] = box.y()
    box_dict['geometry']['width'] = box.size.width()
    box_dict['geometry']['height'] = box.size.height()
    box_dict['geometry']['rotation'] = box.rotation()

    params = {}
    for param in box.get_parameters() :
        current_params = {}
        current_params['type'] = box.get_parameter(param)['type'].__str__()
        current_params['value'] = box.get_parameter(param)['value']

        params[param] = current_params

    box_dict['params'] = params

    return box_dict

def export_input (input : InputConnectionModel) -> Dict :
    """Est la méthode appelée pour créer un dictionnaire contenant les données d'une entrée de box."""

    input_dict = {}

    input_dict['name'] = input.name
    input_dict['isInfinite'] = input.is_infinite
    input_dict['data_type'] = input.data_type.__str__()
    input_dict['text'] = input.text

    # input_dict['color'] = {}
    # input_dict['color']['red'] = input.color.red
    # input_dict['color']['green'] = input.color.green
    # input_dict['color']['blue'] = input.color.blue

    input_dict['signal_index'] = -1

    return input_dict

def export_output (output : OutputConnectionModel) -> Dict : 
    """Est la méthode appelée pour créer un dictionnaire contenant les données d'une sortie de box."""

    output_dict = {}

    output_dict['name'] = output.name
    output_dict['isInfinite'] = output.is_infinite
    # output_dict['count'] = output
    output_dict['data_type'] = output.data_type.__str__()
    output_dict['text'] = output.text

    # output_dict['color'] = {}
    # output_dict['color']['red'] = output.color.red
    # output_dict['color']['green'] = output.color.green
    # output_dict['color']['blue'] = output.color.blue

    output_dict['signal_indices'] = []

    return output_dict

def export_signal (signal_index : int, link_model : AbstractLinkModel) -> Dict :
    """Est la méthode appelée pour créer un dictionnaire contenant les données d'un lien entre des entrées/sorties."""

    signal_dict = {}

    signal_dict['value'] = 0.0
    signal_dict['data_type'] = link_model.data_type.__str__()
    signal_dict['index'] = signal_index
    signal_dict['link_type'] = link_model.link_type.__str__()
    signal_dict['link_value'] = link_model.link_value
    signal_dict['text'] = link_model.text
    # signal_dict['color'] = {}
    # signal_dict['color']['red'] = link_model.color.red
    # signal_dict['color']['green'] = link_model.color.green
    # signal_dict['color']['blue'] = link_model.color.blue
 
    return signal_dict

#-----------------------------------
# Class
#-----------------------------------

class SimulationExporter (AbstractExporter) :
    """Est la classe qui permet d'exporter des fichiers de simulation."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self) :
        
        AbstractExporter.__init__(self)

    # -------------
    # Methods
    # -------------
    def save (self, model: SimulationModel, path : str) -> bool :
        # General informations
        simulation = {}
        simulation['export_version'] = '1.1.0'
        simulation['date_creation'] = model.date_creation
        simulation['date_edition'] = date.today() 
        simulation['version'] = model.version
        simulation['creator'] = model.creator

        # Scheduler parameters
        scheduler = {}
        if model.scheduler_model is None :
            # Set default parameters
            simulation['stop_time'] = 10
            simulation['step_time'] = 0.1
            scheduler['library'] = ''
            scheduler['name'] = ''
        else :
            scheduler['library'] = model.scheduler_model._library
            scheduler['name'] = model.scheduler_model._name

            if model.scheduler_model.params is None : 
                simulation['stop_time'] = 10
                simulation['step_time'] = 0.1
            else :
                simulation['stop_time'] = model.scheduler_model.params.stop_time
                simulation['step_time'] = model.scheduler_model.params.step_time
        simulation['scheduler'] = scheduler
        
        diagram = {}
        boxes = []        
        signals = []

        last_signal_index = 0
        signals_dict : Dict[AbstractSignalModel, int] = {} 

        # Filtering Graphics element
        box_models = [graphic_element for graphic_element in model.graphical_elements if isinstance(graphic_element, BoxModel)]

        # Box Models
        for box_model in box_models:
            # Get dict of box model
            current_box = box_model.to_dict()

            # Inputs
            for inputs_group in box_model.inputs_groups.values(): 
                # Build input dict
                input_group_dict = inputs_group.to_dict()

                for input_index, input_model in enumerate(inputs_group.inputs):
                    # Get input dict
                    input_dict = input_group_dict["inputs"][input_index]

                    # Handle signal
                    if input_model.has_link():
                        # Get link (only one possible)
                        link = input_model._links[0]
                        
                        # If Link already exists
                        if link in signals_dict:
                            input_dict['signal_index'] = signals_dict[link]

                        # If doesn't already exist
                        else: 
                            current_signal = link.to_dict()
                            current_signal['index'] = last_signal_index
                            signals.append(current_signal)

                            signals_dict[link] = last_signal_index
                            input_dict['signal_index'] = last_signal_index
                            last_signal_index += 1

                current_box['inputs_groups'].append(input_group_dict)

            # Outputs
            for outputs_group in box_model.outputs_groups.values(): 
                # Build output dict
                output_group_dict = outputs_group.to_dict()

                for output_index, output_model in enumerate(outputs_group.outputs):
                    # Get output dict
                    output_dict = output_group_dict["outputs"][output_index]
                    output_dict['signal_indices'] = []

                    # Handle signal
                    if output_model.has_link():

                        for link in output_model.links:
                            # If Link already exists
                            if link in signals_dict:
                                output_dict['signal_indices'].append(signals_dict[link])

                            # If doesn't already exist
                            else: 
                                current_signal = link.to_dict()
                                current_signal['index'] = last_signal_index
                                signals.append(current_signal)

                                signals_dict[link] = last_signal_index
                                output_dict['signal_indices'].append(last_signal_index)
                                last_signal_index += 1

                current_box['outputs_groups'].append(output_group_dict)

            # Append box
            boxes.append(current_box)

        diagram['boxes'] = boxes
        diagram['signals'] = signals
        
        simulation['diagram'] = diagram

        serialization = {}
        serialization['simulation'] =  simulation

        with open(path, 'w') as file : 
            yaml.dump(serialization, file)
            return True




#-----------------------------------
# Imports
#-----------------------------------

from typing import Iterable
from PyQt5.QtCore import pyqtSignal  
from PyQt5.QtWidgets import QTabWidget, QWidget
from muphyn.packages.core.utils.log_manager import LogManager
from muphyn.packages.interface.graphical_actions.actions_holder import ActionsHolder

from muphyn.packages.interface.models.editable_models.abstract_editable_model import AbstractEditableModel
from muphyn.packages.interface.models.graphical_models.abstract_graphical_element import AbstractGraphicalElement

#-----------------------------------
# Class
#-----------------------------------

class AbstractEditor (QWidget) : 
    """Est la classe abstraite commune aux éditeurs de contenus."""
    
    # -------------
    # Static Element
    # -------------

    TAB_ID : int = 0

    # -------------
    # Signals
    # -------------
    
    elements_selected_changed = pyqtSignal(object)
    """Est la signal appelé lorsque la sélection contenu dans le modèle est modifié."""

    name_changed = pyqtSignal(object, str)
    """Est le signal appelé lorsque le nom du modèle est modifié."""
    
    # -------------
    # Constructors
    # -------------

    def __init__ (self, tab_holder : QTabWidget, editable_model : AbstractEditableModel, actions_holder : ActionsHolder) :
        QWidget.__init__(self)

        self._tab_holder = tab_holder
        self._editable_model = editable_model
        self._editable_model.name_changed.connect(self.editable_model_name_changed)
        self._editable_model.selection_changed.connect(self.elements_selected_changed)
        self._actions_holder = actions_holder

        # Set Tab ID
        self._id = AbstractEditor.TAB_ID

        # Increment Global Tab ID value
        AbstractEditor.TAB_ID += 1


    # -------------
    # Properties
    # -------------

    @property
    def editable_model (self) -> AbstractEditableModel :
        """Permet de récuperer le model en cours d'édition."""
        return self._editable_model

    @property
    def tab_text (self) -> str :
        """Permet de récuperer le texte à afficher dans le nom de l'onglet."""
        return self._editable_model.name

    @property
    def actions_holder (self) -> ActionsHolder :
        """Permet de récuperer l'action holder de l'éditeur."""
        return self._actions_holder

    
    @property
    def id(self) -> int :
        """ 
        Get the tab ID 

        Returns :
        ---------
        id: int
            ID of the tab
        """
        return self._id

    # -------------
    # Properties
    # -------------

    def editable_model_name_changed (self, kwargs) :
        """Est la méthode appelée lorsque le nom du modèle est modifié."""
        LogManager().set_project_name(self._id, kwargs['name'])
        self.name_changed.emit(self, kwargs['name'])

    def copy (self) -> None :
        """Permet de copier la séléction actuelle dans le presse papier."""
        raise Exception('AbstractEditor.copy is an abstract method and should be overloaded.')

    def cut (self) -> None :
        """Permet de couper la séléction actuelle dans le presse papier."""
        raise Exception('AbstractEditor.cut is an abstract method and should be overloaded.')

    def paste (self) -> None :
        """Permet de coller le contenu du presse papier dans l'éditeur actuel."""
        raise Exception('AbstractEditor.paste is an abstract method and should be overloaded.')

    def selected_elements (self) -> Iterable :
        """Permet de récuperer les éléments sélectionnés dans l'interface."""
        raise Exception('AbstractEditor.selected_elements is an abstract method and should be overloaded.')

    def unslect_elements (self) -> None :
        """Permet de déselectionner les éléments acutellements sélectionnés dans l'interface."""
        raise Exception('AbstractEditor.unslect_elements is an abstract method and should be overloaded.')

    def elements (self) -> Iterable :
        """Permet de récuperer tous les éléments de l'interface."""
        raise Exception('AbstractEditor.elements is an abstract method and should be overloaded.')

    def zoom (self, value : int) -> None :
        """Permet de changer le zoom de l'éditeur."""
        raise Exception('AbstractEditor.zoom is an abstract method and should be overloaded.')
        
    def add_item (self, graphical_element : AbstractGraphicalElement) -> None :
        """Permet d'ajouter un élément graphique à l'interface."""
        raise Exception('AbstractEditor.add_item is an abstract method and should be overloaded.')
        
    def rem_item (self, graphical_element : AbstractGraphicalElement) -> None :
        """Permet de supprimer un élément graphique de l'interface."""
        raise Exception('AbstractEditor.rem_item is an abstract method and should be overloaded.')

    def delete_selection (self) -> None :
        """Permet de supprimer tous les éléments présent dans l'interface."""
        raise Exception('AbstractEditor.delete_selection is an abstract method and should be overloaded.')

    def clear (self) -> None :
        """Permet de supprimer tous les éléments présent dans l'interface."""
        raise Exception('AbstractEditor.clear is an abstract method and should be overloaded.')




#-----------------------------------
# Imports
#-----------------------------------

# Project Imports
from muphyn.packages.core.utils.log_manager import LogManager
from muphyn.packages.interface.graphical_actions.abstract_unique_element_diagram_action import AbstractUniqueElementDiagramAction
from muphyn.packages.interface.models.graphical_models.abstract_graphical_element import AbstractGraphicalElement
from muphyn.packages.interface.models.graphical_models.box_model import BoxModel
from muphyn.packages.interface.models.links_model.abstract_link_model import AbstractLinkModel

#-----------------------------------
# Class
#-----------------------------------

class DiagramChangeElementParamsAction (AbstractUniqueElementDiagramAction) :
    """Est l'action capable de modifier les paramètres d'un élément."""

    # -------------
    # Constructors
    # -------------

    def __init__ (self, graphical_element : AbstractGraphicalElement, param_name : str, new_param_value : object, old_param_value : object) :
        
        AbstractUniqueElementDiagramAction.__init__(self, graphical_element)
        
        self._param_name = param_name
        self._old_param_value = old_param_value
        self._new_param_value = new_param_value

    # -------------
    # Properties
    # -------------

    @property
    def param_name (self) -> str : 
        """Permet de récuperer le nom du paramètre à modifier."""
        return self._param_name

    @property
    def old_param_value (self) -> object :
        """Permet de récuperer la valeur du paramètre avant modification."""
        return self._old_param_value

    @property
    def new_param_value (self) -> object :
        """Permet de récuperer la valeur du paramètre par lequel modifier le nouveau paramètre."""
        return self._new_param_value

    @new_param_value.setter
    def new_param_value (self, new_param_value_ : object) -> None :
        """Permet de modifier la valeur du paramètre par lequel modifier le nouveau paramètre."""

        if new_param_value_ is None : 
            return

        self._new_param_value = new_param_value_

    # -------------
    # Methods
    # -------------

    def do (self) :

        graphical_element = self.graphical_element

        if graphical_element is None : 
            return

        if self.param_name == 'text' :
            graphical_element.text = self.new_param_value 

        elif self.param_name == 'name' :
            graphical_element.name = self.new_param_value

        elif self.param_name == 'color' :
            graphical_element.color = self.new_param_value

        elif self.param_name == 'link_value' :

            if isinstance(graphical_element, AbstractLinkModel) :
                graphical_element.link_value = self.new_param_value
                LogManager().debug(f'-> redo link value : {self.old_param_value} -> {self.new_param_value}')
                graphical_element.self_update()
            else:
                LogManager().error('-> element is not a link')

        elif self.param_name == 'link_type' :
            
            if isinstance(graphical_element, AbstractLinkModel) :
                graphical_element.link_type = self.new_param_value
                LogManager().debug(f'-> redo link value : {self.old_param_value} -> {self.new_param_value}')
                graphical_element.self_update()
            else:
                LogManager().error('-> element is not a link')

        elif self.param_name == 'library' :

            if isinstance(graphical_element, BoxModel) :
                graphical_element.library = self.new_param_value

        elif self.param_name.startswith('params.') :
            
            if isinstance(graphical_element, BoxModel) : 
                graphical_element.set_parameter(self.param_name.split('.')[1], self.new_param_value)
        
        graphical_element.update(graphical_element.boundingRect())

    def undo (self) :

        graphical_element = self.graphical_element

        if graphical_element is None : 
            return
                
        if self.param_name == 'text' :
            graphical_element.text = self.old_param_value 

        elif self.param_name == 'name' :
            graphical_element.name = self.old_param_value

        elif self.param_name == 'color' :
            graphical_element.color = self.old_param_value

        elif self.param_name == 'link_value' :
            
            if isinstance(graphical_element, AbstractLinkModel) :
                graphical_element.link_value = self.old_param_value
                LogManager().debug(f'-> redo link value : {self.new_param_value} -> {self.old_param_value}')
                graphical_element.self_update()

        elif self.param_name == 'link_type' :
            
            if isinstance(graphical_element, AbstractLinkModel) :
                graphical_element.link_type = self.old_param_value
                LogManager().debug(f'-> redo link value : {self.new_param_value} -> {self.old_param_value}')
                graphical_element.self_update()

        elif self.param_name == 'library' :

            if isinstance(graphical_element, BoxModel) :
                graphical_element.library = self.old_param_value

        elif self.param_name.startswith('params.') :
            
            if isinstance(graphical_element, BoxModel) : 
                graphical_element.set_parameter(self.param_name.split('.')[1], self.old_param_value)

        graphical_element.update(graphical_element.boundingRect())


import datetime
import logging, logging.handlers, logging.config
import os
import platform
from threading import Lock

from muphyn.packages.core.utils.file import CreateDirectory
from muphyn.packages.core.utils.singleton import SingletonMetaClass

# class LogManagerMeta(type):
#     """
#     This is a thread-safe implementation of Singleton.
#     """

#     _instances = {}

#     _lock: Lock = Lock()
#     """
#     We now have a lock object that will be used to synchronize threads during
#     first access to the Singleton.
#     """

#     def __call__(cls, *args, **kwargs):
#         with cls._lock:
#             if cls not in cls._instances:
#                 instance = super().__call__(*args, **kwargs)
#                 cls._instances[cls] = instance
#         return cls._instances[cls]


class LogManager(metaclass=SingletonMetaClass):

    DisabledLoggers = ['PIL.Image', 'PIL', 'PIL.PngImagePlugin', 'matplotlib.ticker', 'matplotlib', 'matplotlib.dates', 'matplotlib.artist', 'matplotlib.lines', 'matplotlib.dviread', 'matplotlib.afm', 'matplotlib.font_manager', 'matplotlib.mathtext', 'matplotlib.textpath', 'matplotlib.backend_bases', 'matplotlib.text', 'matplotlib.colorbar', 'matplotlib.image', 'matplotlib.style.core', 'matplotlib.style', 'matplotlib.category', 
'matplotlib.axis', 'matplotlib.axes._base', 'matplotlib.axes', 'matplotlib.axes._axes', 'matplotlib.gridspec', 'matplotlib.figure', 'matplotlib.pyplot', 'opcua.ua.uatypes', 'opcua.ua', 'opcua', 'concurrent.futures', 'concurrent', 'asyncio', '__name__', 'opcua.uaprotocol', 'opcua.common.copy_node', 'opcua.common', 'opcua.common.instantiate', 'opcua.common.structures', 'opcua.client.client', 'opcua.client', 'opcua.server.binary_server_asyncio', 'opcua.server', 'opcua.server.server', 'OMPython', 'opcua.client.ua_client', 'opcua.client.ua_client.Socket', 'matplotlib.backends.qt_editor._formlayout', 'matplotlib.backends.qt_editor', 'matplotlib.backends', 'matplotlib.legend']

    def __init__(self, log_level: int = logging.INFO) -> None:
        logging.basicConfig(format="%(asctime)s %(levelname)s: %(message)s")

        # Set logging level
        logging.root.setLevel(log_level)

        # Disable matplotlib font manager debug
        for logger_name in LogManager.DisabledLoggers:
            logging.getLogger(logger_name).disabled = True

        # Init selected project
        self._current_project_id = -1

        # Create Global Log File Directory
        if platform.system() == "Windows":
            global_log_file_directory = os.path.join(os.getenv("APPDATA"), "Muphyn", "LOGS")
        else:
            global_log_file_directory = os.path.join(os.getcwd(), "LOGS")
        CreateDirectory(global_log_file_directory)

        # Build Global Log File
        global_log_file_path = os.path.join(global_log_file_directory, self._build_log_filename())

        # Create Global logger
        global_logger_name = self._create_logger(-1, "", global_log_file_path, logger_name="main_logger")

        # Init Tab ID 
        self._project_log_files = {
            self._current_project_id: {"ProjectName": "Global", "LoggerName": global_logger_name}
        }

    
    # ----------------
    # Logging Methods
    # ----------------

    def _log_message(self, message: str, level: int, logger_name: str):
        """
        Private Function: It writes in the selected log file the log message

        Parameters :
        ------------
        - message: str
            Log message to write in the file

        - level: int
            Level of the log message
        
        - log_file: str
            Target File Path for log message
        """
        # Get Logger
        logger = logging.getLogger(logger_name)

        # Log message
        logger.log(level, message)

    def _log_global_message(self, message: str, level: int):
        """
        Private Function: It writes in the global log file the log message

        Parameters :
        ------------
        - message: str
            Log message to write in the file

        - level: int
            Level of the log message
        """
        # Get Global Loger Name
        logger_name = self._project_log_files[-1]["LoggerName"]

        # Current Project Name
        current_project_name = self._project_log_files[self._current_project_id]["ProjectName"]

        # Log Message
        self._log_message(f"({current_project_name}) {message}", level, logger_name)

    def _log_current_project(self, message: str, level:int):
        """
        Private Function: It writes in the current project log file the log message.
            If there is no project opened or log file created for the current project 
            it writes into the global log file

        Parameters :
        ------------
        - message: str
            Log message to write in the file

        - level: int
            Level of the log message
        """
        if self._current_project_id != -1 and self._current_project_id in self._project_log_files:
            # Get logger name
            project_logger_name = self._project_log_files[self._current_project_id]["LoggerName"]

            # Log Message
            self._log_message(message, level, project_logger_name)
        else:
            # Log message in global Log File
            self._log_global_message(message, level)

    def log(self, message: str, level: int, is_global_message=False):
        """
        Public Function: It writes a log message in a Log File. 
            The target Log file can be selected between global log file or current project log file

        Parameters :
        ------------
        - message: str
            Log message to write in the file

        - level: int
            Level of the log message. Possible values : 
                * logging.DEBUG
                * logging.ERROR
                * logging.INFO
                * logging.WARNING

        - is_global_message: bool (optional)
            Select if the log message is written in the current project log file or global log file
        """
        # Log in global log file
        self._log_global_message(message, level)

        # Log in project log file
        if not is_global_message:
            self._log_current_project(message, level)

    def critical(self, message: str, is_global_message=False):
        """
        Public Function: It writes a critical message in the target Log File

        Parameters :
        ------------
        - message: str
            Log message to write in the file

        - is_global_message: bool (optional)
            Select if the log message is written in the current project log file or global log file
        """
        self.log(message, logging.CRITICAL, is_global_message)

    def debug(self, message: str, is_global_message=False):
        """
        Public Function: It writes a debug message in the target Log File

        Parameters :
        ------------
        - message: str
            Log message to write in the file

        - is_global_message: bool (optional)
            Select if the log message is written in the current project log file or global log file
        """
        self.log(message, logging.DEBUG, is_global_message)

    def error(self, message: str, is_global_message=False):
        """
        Public Function: It writes an error message in the target Log File

        Parameters :
        ------------
        - message: str
            Log message to write in the file

        - is_global_message: bool (optional)
            Select if the log message is written in the current project log file or global log file
        """
        self.log(message, logging.ERROR, is_global_message)

    def info(self, message: str, is_global_message=False):
        """
        Public Function: It writes an information message in the target Log File

        Parameters :
        ------------
        - message: str
            Log message to write in the file

        - is_global_message: bool (optional)
            Select if the log message is written in the current project log file or global log file
        """
        self.log(message, logging.INFO, is_global_message)

    def warning(self, message: str, is_global_message=False):
        """
        Public Function: It writes a warning message in the target Log File

        Parameters :
        ------------
        - message: str
            Log message to write in the file

        - is_global_message: bool (optional)
            Select if the log message is written in the current project log file or global log file
        """
        self.log(message, logging.WARNING, is_global_message)

    def setLevel(self, logging_level:int):
        logging.root.setLevel(logging_level)

    
    # ----------------
    # Other Methods
    # ----------------

    def _build_log_filename(self, project_name: str = None):
        """
        Private Function: Build the log file name based on the time of creation
            and the project name if this one is given

        Parameters :
        ------------
        - project_name: str (optionnal)
            Project name related to the log file

        Returns :
        ---------
        str
            Built log file name
        """
        # Get Time
        now = datetime.datetime.now()

        # Build File Name
        log_filename = now.strftime("%Y-%m-%d_%H-%M-%S")+".log"

        if project_name is not None:
            log_filename = f"{project_name}_{log_filename}"

        return log_filename

    def _create_logger(self, tab_id: int, project_name: str, project_log_file_path: str, logger_name:str = None):
        """
        Private Function: Create a logging.Logger object based on project information
        
        Parameters :
        ------------
        - tab_id: int
            ID of the project tab

        - project_name: str
            Project name

        - project_directory: str
            Project directory path

        - logger_name: str (optionnal)
            Logger name. The default logger name is equal to "{tab_id}_{project_name}"

        Returns :
        ---------
        str
            Name of the previously created logger object.
            This name is used to access the logger by using the function logging.getLogger(logger_name)

        """
        # Build Logger Name
        if logger_name is None:
            logger_name = f"{tab_id}_{project_name}" 

        # Get Logger if it exists or Build Logger
        logger = logging.getLogger(logger_name)
        # logger.setLevel(LogManager.LoggingLevel)

        # Remove all existing handler if it exists
        for handler in logger.handlers:
            logger.removeHandler(handler)
        
        # Build formatter
        formatter = logging.Formatter("%(asctime)s %(levelname)s: %(message)s")

        # Build File Handler
        file_handler = logging.FileHandler(project_log_file_path, mode="w", encoding="utf-8")
        file_handler.setFormatter(formatter)

        # Append handler
        logger.addHandler(file_handler)

        return logger_name

    def add_project_logger(self, tab_id: int, project_name: str, project_directory: str):
        """
        Public Function: Append the project general informations (tab ID, project name & project directory path)

        Parameters :
        ------------
        - tab_id: int
            ID of the project tab

        - project_name: str
            Project name

        - project_directory: str
            Project directory path

        """
        if tab_id in self._project_log_files:
            # Write information in global LOG File
            self.info("Trying to add project information that already exists", is_global_message=True)
        else:
            # Create Log Directory
            project_log_directory = os.path.join(project_directory, "LOGS")
            CreateDirectory(project_log_directory)

            # Build Log File path
            project_log_file_path = os.path.join(project_log_directory, self._build_log_filename(project_name))

            # Create Logger
            logger_name = self._create_logger(tab_id, project_name, project_log_file_path)

            # Append Project Log File Informations
            self._project_log_files[tab_id] = {"ProjectName": project_name, "LoggerName": logger_name}

    def remove_project_logger(self, tab_id):
        """
        Public Function: Remove logger related a project

        Parameters :
        ------------
        - tab_id: int
            ID of the project tab
        """
        if tab_id in self._project_log_files:
            # Get Project Name
            logger_name = self._project_log_files[tab_id]["LoggerName"]

            # Remove from dictionnary
            del self._project_log_files[tab_id]
            
            if logger_name in logging.Logger.manager.loggerDict:
                # Remove logger
                del logging.Logger.manager.loggerDict[logger_name]

    def set_project_name(self, tab_id: int, project_name: str):
        """
        Public Function: Set logger path by deleting the existing logger and recreating a new logger

        Parameters :
        ------------
        - tab_id: int
            ID of the project tab
        - project_name: str
            New project name
        """
        if tab_id in self._project_log_files:
            self._project_log_files[tab_id]["ProjectName"] = project_name

    def set_project_path(self, tab_id: int, project_path: str):
        """
        Public Function: Set logger path by deleting the existing logger and recreating a new logger

        Parameters :
        ------------
        - tab_id: int
            ID of the project tab
        - project_path: str
            New project path
        """
        if tab_id in self._project_log_files:
            # Get Project Name
            project_name = self._project_log_files[tab_id]["ProjectName"]

            # Remove Existing Logger
            self.remove_project_logger(tab_id)

            # Create New Logger
            self._create_logger(tab_id, project_name, project_path)

    def set_current_project_id(self, tab_id: int):
        """
        Public Function: Set the current active project

        Parameters :
        ------------
        - tab_id: int
            ID of the project tab
        """
        if tab_id in self._project_log_files:
            # Set the current project id
            self._current_project_id = tab_id
        else:
            # Reset the current project id
            self._current_project_id = -1

    def reset_current_project_id(self):
        """
        Public Function: Reset the current active project
        """
        self.set_current_project_id(-1)

    def get_all_logger(self) -> list[str]:
        return [name for name in logging.root.manager.loggerDict]


from threading import Lock

class SingletonMetaClass(type):
    """
    This is a thread-safe implementation of Singleton.
    """

    _instances = {}

    _lock: Lock = Lock()
    """
    We now have a lock object that will be used to synchronize threads during
    first access to the Singleton.
    """

    def __call__(cls, *args, **kwargs):
        with cls._lock:
            if cls not in cls._instances:
                instance = super().__call__(*args, **kwargs)
                cls._instances[cls] = instance
        return cls._instances[cls]